from __future__ import annotations

import argparse
import numbers
import re
import warnings
from gettext import gettext as _, ngettext
from pathlib import Path
from string import printable
from textwrap import dedent
from typing import TYPE_CHECKING, Any

from streamlink import __version__ as streamlink_version, logger
from streamlink.exceptions import StreamlinkDeprecationWarning
from streamlink.logger import getLogger
from streamlink.options import Options
from streamlink.utils.args import boolean, comma_list, comma_list_filter, filesize, keyvalue, num
from streamlink.utils.times import hours_minutes_seconds_float
from streamlink_cli.constants import STREAM_PASSTHROUGH
from streamlink_cli.exceptions import StreamlinkCLIError
from streamlink_cli.output.player import PlayerArgs, PlayerOutput
from streamlink_cli.utils import find_default_player


if TYPE_CHECKING:
    from collections.abc import Callable

    from streamlink.plugin import Plugin
    from streamlink.session import Streamlink
    from streamlink.user_input import UserInputRequester


log = getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    # noinspection PyUnresolvedReferences,PyProtectedMember
    NESTED_ARGUMENT_GROUPS: dict[argparse._ArgumentGroup | None, list[argparse._ArgumentGroup]]

    _RE_PRINTABLE = re.compile(rf"[{re.escape(printable)}]")
    _RE_OPTION = re.compile(r"^(?P<name>[A-Za-z0-9-]+)(?:(?P<op>\s*=\s*|\s+)(?P<value>.*))?$")

    def __init__(self, *args, **kwargs):
        self.NESTED_ARGUMENT_GROUPS = {}
        self.color = True  # pre 3.14 compat
        super().__init__(*args, **kwargs)

    # noinspection PyUnresolvedReferences,PyProtectedMember
    def add_argument_group(
        self,
        *args,
        parent: argparse._ArgumentGroup | None = None,
        **kwargs,
    ) -> argparse._ArgumentGroup:
        group = super().add_argument_group(*args, **kwargs)
        if parent not in self.NESTED_ARGUMENT_GROUPS:
            self.NESTED_ARGUMENT_GROUPS[parent] = [group]
        else:
            self.NESTED_ARGUMENT_GROUPS[parent].append(group)
        return group

    def convert_arg_line_to_args(self, arg_line: str) -> list[str]:
        # Strip any non-printable characters that might be in the
        # beginning of the line (e.g. Unicode BOM marker).
        match = self._RE_PRINTABLE.search(arg_line)
        if not match:
            return []
        arg_line = arg_line[match.start() :].strip()

        # Skip lines that do not start with a valid option (e.g. comments)
        option = self._RE_OPTION.match(arg_line)
        if not option:
            return []

        name, op, value = option.group("name", "op", "value")
        prefix = self.prefix_chars[0] if len(name) == 1 else self.prefix_chars[0] * 2
        if value or op:
            return [f"{prefix}{name}={value}"]
        else:
            return [f"{prefix}{name}"]

    def _match_argument(self, action: argparse.Action, arg_strings_pattern: str) -> int:
        # - https://github.com/streamlink/streamlink/issues/971
        # - https://bugs.python.org/issue9334
        # - https://github.com/python/cpython/blame/v3.13.0rc2/Lib/argparse.py#L2227-L2247

        # match the pattern for this action to the arg strings
        nargs_pattern = self._get_nargs_pattern(action)
        match = re.match(nargs_pattern, arg_strings_pattern)

        # if no match, see if we can emulate optparse and return the
        # required number of arguments regardless of their values
        if match is None:
            nargs = action.nargs if action.nargs is not None else 1
            if isinstance(nargs, numbers.Number) and len(arg_strings_pattern) >= int(nargs):
                return int(nargs)

        # raise an exception if we weren't able to find a match
        if match is None:
            if isinstance(action.nargs, numbers.Number):
                msg = ngettext("expected %s argument", "expected %s arguments", int(action.nargs)) % int(action.nargs)
            else:
                nargs_errors: dict[str | int | None, str] = {
                    argparse.OPTIONAL: _("expected at most one argument"),
                    argparse.ONE_OR_MORE: _("expected at least one argument"),
                }
                msg = nargs_errors.get(action.nargs, _("expected one argument"))
            raise argparse.ArgumentError(action, msg)

        # return the number of arguments matched
        return len(match.group(1))

    # disable color output for the "usage" text
    def format_usage(self):
        color = self.color
        self.color = False
        try:
            return super().format_usage()
        finally:
            self.color = color

    # fix `--help` not including nested argument groups
    def format_help(self):
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)

        def format_group(parent):
            if parent not in self.NESTED_ARGUMENT_GROUPS:
                return
            # positionals, optionals and user-defined groups
            for action_group in self.NESTED_ARGUMENT_GROUPS[parent]:
                formatter.start_section(action_group.title)
                formatter.add_text(action_group.description)
                formatter.add_arguments(action_group._group_actions)
                format_group(action_group)
                formatter.end_section()

        format_group(None)

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """A nicer help formatter.

    Help for arguments can be indented and contain new lines.
    It will be de-dented and arguments in the help will be
    separated by a blank line for better readability.

    Originally written by Jakub Roztocil of the httpie project.
    """

    def __init__(self, max_help_position=4, *args, **kwargs):
        # A smaller indent for args help.
        kwargs["max_help_position"] = max_help_position
        super().__init__(*args, **kwargs)

    def _split_lines(self, text, width):
        return f"{dedent(text).strip()}\n\n".splitlines()


def build_parser():
    parser = ArgumentParser(
        prog="streamlink",
        fromfile_prefix_chars="@",
        formatter_class=HelpFormatter,
        add_help=False,
        usage="%(prog)s [OPTIONS] <URL> [STREAM]",
        description=dedent("""
            Streamlink is a command-line utility that extracts streams from various
            services and pipes them into a video player of choice.
        """),
        epilog=dedent("""
            For more in-depth documentation see:
              https://streamlink.github.io/

            Please report broken plugins or bugs to the issue tracker on GitHub:
              https://github.com/streamlink/streamlink/issues
        """),
    )

    positional = parser.add_argument_group("Positional arguments")
    positional.add_argument(
        "url",
        metavar="URL",
        nargs="?",
        help="""
            A URL to attempt to extract streams from.

            Usually, the protocol of http(s) URLs can be omitted (`https://`),
            depending on the implementation of the plugin being used.

            Alternatively, the URL can also be specified by using the --url option.
        """,
    )
    positional.add_argument(
        "stream",
        metavar="STREAM",
        nargs="?",
        type=comma_list,
        help="""
            Stream to play.

            Use `best` or `worst` for selecting the highest or lowest available quality.

            Fallback streams can be specified by using a comma-separated list:

              "720p,480p,best"

            If no stream is specified and --default-stream is not used, then a list of available streams will be printed.
        """,
    )

    general = parser.add_argument_group("General options")
    general.add_argument(
        "-h",
        "--help",
        action="store_true",
        help="""
            Show this help message and exit.
        """,
    )
    general.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {streamlink_version}",
        help="""
            Show version string and exit.
        """,
    )
    general.add_argument(
        "--version-check",
        action="store_true",
        help="""
            Run a version check and exit.
        """,
    )
    general.add_argument(
        "--auto-version-check",
        type=boolean,
        metavar="{yes,true,1,on,no,false,0,off}",
        default=False,
        help="""
            Enable or disable the automatic check for a new version of Streamlink.

            Default is "no".
        """,
    )
    general.add_argument(
        "--config",
        action="append",
        metavar="FILENAME",
        help="""
            Load options from this config file.

            Can be repeated to load multiple files, in which case the options are
            merged on top of each other where the last config has highest priority.
        """,
    )
    general.add_argument(
        "--no-config",
        action="store_true",
        help="""
            Disable loading any default or custom config files.
        """,
    )
    general.add_argument(
        "--locale",
        type=str,
        metavar="LOCALE",
        help="""
            Override the system's locale setting, for selecting the preferred subtitle and audio language.

            The locale is formatted as `[language_code]_[country_code]`, e.g. `en_US` or `es_ES`.

            Default is system locale.
        """,
    )

    logging = parser.add_argument_group("Logging arguments")
    logging.add_argument(
        "-l",
        "--loglevel",
        metavar="LEVEL",
        choices=logger.levels,
        default="info",
        help=f"""
            Set the log message threshold.

            Valid levels are, in order of increasing verbosity:

            {", ".join([f"`{level}`" for level in logger.levels])}

            Default is "info".
        """,
    )
    logging.add_argument(
        "--logformat",
        metavar="FORMAT",
        help="""
            Set a custom logging format.

            See the Python standard library's `logging.Formatter` docs for more information about the logging format
            and the available `LogRecord` attributes. Streamlink's formatter uses the curly brace style.

            The default format depends on the chosen log level.

            For verbose levels (`trace` and `all`):

            Default is "[{asctime}][{threadName}][{name}][{levelname}] {message}".

            Otherwise:

            Default is "[{name}][{levelname}] {message}".
        """,
    )
    logging.add_argument(
        "--logdateformat",
        metavar="DATEFORMAT",
        help="""
            Set a custom logging date format.

            This formats the `LogRecord`'s `asctime` attribute via `strftime()`.

            The default date format depends on the chosen log level.

            For verbose levels (`trace` and `all`):

            Default is "%%H:%%M:%%S.%%f".

            Otherwise:

            Default is "%%H:%%M:%%S".
        """,
    )
    logging.add_argument(
        "--logfile",
        metavar="FILE",
        help="""
            Append log output to `FILE` instead of writing to stdout/stderr.

            User prompts and download progress won't be written to `FILE`.

            A value of `-` (dash) will set the file name to an ISO8601-like string
            and will choose the following default log directories.

            Windows:

              %%TEMP%%\\streamlink\\logs

            macOS:

              ${HOME}/Library/Logs/streamlink

            Linux/BSD:

              ${XDG_STATE_HOME:-${HOME}/.local/state}/streamlink/logs
        """,
    )
    logging.add_argument(
        "-Q",
        "--quiet",
        action="store_true",
        help="""
            Suppress all console and log output, and also disable user prompts.
        """,
    )
    logging.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="""
            Output JSON representations instead of the normal text output.

            Useful for external scripting.
        """,
    )

    plugin = parser.add_argument_group("Plugin options")
    plugin.add_argument(
        "--plugins",
        action="store_true",
        help="""
            Print a list of all currently installed plugins.
        """,
    )
    plugin.add_argument(
        "--show-matchers",
        metavar="PLUGIN",
        help="""
            Show the list of matchers of a specific plugin (URL regex pattern with opt. priority and opt. name).

            The output is a human-readable pseudo YAML format. Please use --json when reading matcher data programmatically.
        """,
    )
    plugin.add_argument(
        "--can-handle-url",
        metavar="URL",
        help="""
            Check if Streamlink has a plugin that can handle the specified URL.

            Status code is `0` on success, `1` on failure.

            Useful for external scripting.
        """,
    )
    plugin.add_argument(
        "--can-handle-url-no-redirect",
        metavar="URL",
        help="""
            Same as --can-handle-url, but without following redirects when looking up the URL.
        """,
    )
    plugin.add_argument(
        "--no-plugin-cache",
        action="store_true",
        default=None,
        help="""
            Disable I/O of the plugin key-value store.

            If disabled, plugins won't be able to load or store data like cookies, authentication data, etc.
            The data which is loaded or stored depends on each plugin implementation.
        """,
    )
    plugin.add_argument(
        "--no-plugin-sideloading",
        action="store_true",
        help="""
            Disable automatic sideloading of third-party plugins from the default location.

            See the plugin-sideloading documentation for where third-party plugins are loaded from.
        """,
    )
    plugin.add_argument(
        "--plugin-dir",
        dest="plugin_dirs",
        metavar="DIRECTORY",
        action="append",
        help="""
            Load additional plugins from this directory.

            Can be set multiple times to load plugins from multiple directories.
        """,
    )
    plugin.add_argument(
        "--plugin-dirs",
        metavar="DIRECTORY",
        type=comma_list,
        action="extend",
        help="""
            Load additional plugins from a list of comma-separated directories. (deprecated)
        """,
    )

    network = parser.add_argument_group("Network arguments")
    network.add_argument(
        "--interface",
        type=str,
        metavar="INTERFACE",
        help="""
            Select the network interface, either by IP address or by interface name.

            The following optional prefixes are supported:

            - `if!<name>` interface name
            - `host!<name>` IP address or hostname
            - `ifhost!<interface>!<host>` interface name and IP address or hostname

            Note: On Windows, only IP addresses are supported and prefixes are ignored.
        """,
    )
    network.add_argument(
        "-4",
        "--ipv4",
        action="store_true",
        default=None,
        help="""
            Resolve address names to IPv4 only. This option overrides --ipv6.
        """,
    )
    network.add_argument(
        "-6",
        "--ipv6",
        action="store_true",
        default=None,
        help="""
            Resolve address names to IPv6 only. This option overrides --ipv4.
        """,
    )

    player = parser.add_argument_group("Player options")
    player.add_argument(
        "-p",
        "--player",
        metavar="PATH",
        type=Path,
        default=find_default_player(),
        help="""
            Set the player executable that will be launched (unless a different output method was chosen).

            Either set an absolute or relative path to the player executable, or just set the executable's name
            if it can be resolved from the paths of the system's `PATH` environment variable.

            In addition to setting the player executable path, custom player arguments can be set via --player-args.

            Note: In the past, --player allowed defining additional player arguments, which as a consequence required wrapping
            player paths that contained spaces in quotation marks. This is unsupported since release `6.0.0`.

            Default is VLC player, if available.
        """,
    )
    player.add_argument(
        "-a",
        "--player-args",
        metavar="ARGUMENTS",
        default="",
        help=f"""
            Set a string of custom --player launch arguments that will be parsed and tokenized.

            The value can contain formatting variables surrounded by curly braces, `{{` and `}}`.
            Curly brace characters can be escaped by doubling, e.g. `{{{{` and `}}}}`.

            Available formatting variables:

            `{{{PlayerOutput.PLAYER_ARGS_INPUT}}}`
                This is the input argument that the --player will receive. For standard input (stdin),
                it is `-` (dash), but it can also be a file path or URL, depending on the options used.
                If unset, then the player input argument will be appended to the parsed player arguments list.

            `{{{PlayerOutput.PLAYER_ARGS_TITLE}}}`
                The automatically generated player title arguments, if a supported --player was found. See --title for more.
                If unset, automatically generated player title arguments will be prepended to the parsed player arguments list.

            Example:

              %(prog)s -p vlc -a "--play-and-exit --no-one-instance" <url> [stream]

            Default is "".
        """,
    )
    player.add_argument(
        "--player-env",
        metavar="KEY=VALUE",
        type=keyvalue,
        action="append",
        help="""
            Add an additional environment variable to the spawned --player process, in addition to the ones inherited from
            the Streamlink/Python parent process. This allows setting player environment variables in config files.

            Can be repeated to add multiple environment variables.
        """,
    )
    player.add_argument(
        "-v",
        "--player-verbose",
        action="store_true",
        help="""
            Write the --player's stdout/stderr output to Streamlink's stdout/stderr output.
        """,
    )
    player.add_argument(
        "--verbose-player",
        dest="player_verbose",
        action="store_true",
        help="""
            Deprecated in favor of --player-verbose.
        """,
    )
    player.add_argument(
        "-n",
        "--player-fifo",
        action="store_true",
        help="""
            Make the --player read the stream through a named pipe instead of the stdin pipe.
        """,
    )
    player.add_argument(
        "--fifo",
        dest="player_fifo",
        action="store_true",
        help="""
            Deprecated in favor of --player-fifo.
        """,
    )
    player.add_argument(
        "--player-http",
        action="store_true",
        help="""
            Make the --player read the stream through HTTP instead of the stdin pipe.
        """,
    )
    player.add_argument(
        "--player-continuous-http",
        action="store_true",
        help="""
            Make the --player read the stream through HTTP, but unlike --player-http,
            it will continuously try to open the stream if the player requests it.

            This enables the handling of stream disconnects if the player is
            capable of reconnecting to a HTTP stream. This is usually done by
            setting the player to a "repeat mode".
        """,
    )
    player.add_argument(
        "--player-external-http",
        action="store_true",
        help="""
            Serve stream data through HTTP without opening the --player. This is
            useful to allow external devices like smartphones or streaming boxes to
            watch streams they wouldn't be able to otherwise.

            The default behavior is similar to the --player-continuous-http option,
            but no player program will be started, and the server will listen on all available
            connections instead of just in the local (loopback) interface.

            See --player-external-http-interface for choosing a specific network interface, and
            see --player-external-http-port for choosing a non-randomized port.

            Optionally, the --player-external-http-continuous option allows for disabling
            the continuous run-mode, so that Streamlink will stop when the stream ends.

            The URLs that can be used to access the stream will be printed to the
            console, and the server can be interrupted using CTRL-C.
        """,
    )
    player.add_argument(
        "--player-external-http-continuous",
        type=boolean,
        metavar="{yes,true,1,on,no,false,0,off}",
        default=True,
        help="""
            Set the run-mode of --player-external-http to continuous or non-continuous.

            In the continuous run-mode, Streamlink will keep running after the stream has ended
            and will wait for the next HTTP request being made unless it gets shut down via CTRL-C.

            If set to non-continuous, Streamlink will stop once the stream has ended.

            Default is true.
        """,
    )
    player.add_argument(
        "--player-external-http-interface",
        metavar="INTERFACE",
        help="""
            Set the network interface on which the HTTP server will be listening on.
            If unset or set to `0.0.0.0`, all available interfaces will be bound.
        """,
    )
    player.add_argument(
        "--player-external-http-port",
        metavar="PORT",
        type=num(int, ge=0, le=65535),
        default=0,
        help="""
            Set the port of the external HTTP server if that mode is enabled.
            Omit or set to `0` to use a random high ( >1024) port.
        """,
    )
    player.add_argument(
        "--player-passthrough",
        metavar="TYPES",
        type=comma_list_filter(STREAM_PASSTHROUGH),
        default=[],
        help=f"""
            A comma-delimited list of stream types to pass to the --player as a URL to
            let it handle the transport of the stream instead of Streamlink.

            Stream types that can be converted into a playable URL are:

            {", ".join(STREAM_PASSTHROUGH)}

            Make sure the player can handle the stream type when using this.
        """,
    )
    player.add_argument(
        "--player-no-close",
        action="store_true",
        help="""
            By default, Streamlink will close the --player when the stream ends.
            This is to avoid "dead" GUI players lingering after Streamlink has exited.

            It does however have the side-effect of sometimes closing a
            player before it has played back all of its cached data.

            This option will instead let the player decide when to exit.
        """,
    )
    # noinspection PyTypeChecker
    player.add_argument(
        "-t",
        "--title",
        metavar="TITLE",
        help=f"""
            Change the title of the --player's window.

            Please see the "Metadata variables" section of Streamlink's CLI documentation for all available metadata variables,
            as well as the "Plugins" section for the list of metadata variables defined in each plugin.

            Only the following players are supported:
            {"".join(f"{chr(0x0A)}            - {pn}" for pn in PlayerArgs.get_player_names())}

            Example:

                %(prog)s -p mpv --title "{{author}} - {{category}} - {{title}}" <URL> [STREAM]
        """,
    )

    output = parser.add_argument_group("File output options")
    output.add_argument(
        "-O",
        "--stdout",
        action="store_true",
        help="""
            Write stream data to `stdout` instead of playing it in the --player.
        """,
    )
    output.add_argument(
        "-o",
        "--output",
        metavar="FILENAME",
        help="""
            Write stream data to `FILENAME` instead of playing it in the --player.
            If `FILENAME` is set to `-` (dash), then the stream data will be written to `stdout`,
            similar to the --stdout argument.

            Directories and subdirectories will be created if they do not exist, if filesystem permissions allow.

            Unless --force is set, Streamlink will ask for confirmation before writing if `FILENAME` already exists.

            Please see the "Metadata variables" section of Streamlink's CLI documentation for all available metadata variables,
            as well as the "Plugins" section for the list of metadata variables defined in each plugin.

            Unsupported characters in substituted variables will be replaced with an underscore.

            Example:

                %(prog)s --output "~/recordings/{author}/{category}/{id}-{time:%%Y%%m%%d%%H%%M%%S}.ts" <URL> [STREAM]
        """,
    )
    output.add_argument(
        "-r",
        "--record",
        metavar="FILENAME",
        help="""
            Write stream data to `FILENAME` while at the same time allowing playback in the --player or writing it to --stdout.
            If `FILENAME` is set to `-` (dash), then the stream data will be written to `stdout`,
            similar to the --stdout argument, while still opening the player.

            Directories and subdirectories will be created if they do not exist, if filesystem permissions allow.

            Unless --force is set, Streamlink will ask for confirmation before writing if `FILENAME` already exists.

            Please see the "Metadata variables" section of Streamlink's CLI documentation for all available metadata variables,
            as well as the "Plugins" section for the list of metadata variables defined in each plugin.

            Unsupported characters in substituted variables will be replaced with an underscore.

            Example:

                %(prog)s --record "~/recordings/{author}/{category}/{id}-{time:%%Y%%m%%d%%H%%M%%S}.ts" <URL> [STREAM]
        """,
    )
    output.add_argument(
        "-R",
        "--record-and-pipe",
        metavar="FILENAME",
        help="""
            Deprecated in favor of --stdout --record=FILENAME.
        """,
    )
    output.add_argument(
        "--fs-safe-rules",
        metavar="{POSIX,Windows}",
        choices=["POSIX", "Windows"],
        type=str,
        help="""
            The rules used to make formatting variables filesystem-safe are chosen
            automatically according to the type of system in use. This overrides
            the automatic detection.

            Intended for use when Streamlink is running on a UNIX-like OS but writing
            to Windows filesystems such as NTFS; USB devices using VFAT or exFAT; CIFS
            shares that are enforcing Windows filename limitations, etc.

            These characters are replaced with an underscore for the rules in use:

            - POSIX: `\\x00-\\x1F /`
            - Windows: `\\x00-\\x1F \\x7F " * / : < > ? \\ |`
        """,
    )
    output.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="""
            When using --output or --record, always write to file even if it already exists (overwrite).
        """,
    )
    output.add_argument(
        "--skip",
        action="store_true",
        help="""
            When using --output or --record, never write to file if it already exists (don't prompt).

            Takes precedence over --force.
        """,
    )
    output.add_argument(
        "--progress",
        metavar="{yes,force,no}",
        choices=("yes", "force", "no"),
        default="yes",
        help="""
            When using --output or --record, show or hide the download progress bar, or force it if there's no terminal.

            Default is yes.
        """,
    )

    stream = parser.add_argument_group("Stream options")
    stream.add_argument(
        "--url",
        dest="url_param",
        metavar="URL",
        help="""
            A URL to attempt to extract streams from.

            Usually, the protocol of http(s) URLs can be omitted (`https://`),
            depending on the implementation of the plugin being used.

            This is an alternative to setting the URL using a positional argument and can be useful if set in a config file.
        """,
    )
    stream.add_argument(
        "--default-stream",
        type=comma_list,
        metavar="STREAM",
        help="""
            Stream to play.

            Use `best` or `worst` for selecting the highest or lowest available quality.

            Fallback streams can be specified by using a comma-separated list:

              "720p,480p,best"

            This is an alternative to setting the stream using a positional argument and can be useful if set in a config file.
        """,
    )
    stream.add_argument(
        "--stream-url",
        action="store_true",
        help="""
            If possible, translate the resolved stream to a URL and print it.
        """,
    )
    stream.add_argument(
        "--retry-streams",
        metavar="DELAY",
        type=num(float, ge=0),
        help="""
            Retry fetching the list of available streams until streams are found
            while waiting `DELAY` second(s) between each attempt.

            The number of retry attempts can be capped with --retry-max.
            A default value of ``1`` is implied for non-zero values of --retry-max.

            If both --retry-streams and --retry-max are set to `0`, then only one attempt will be made
            to fetch the list of available streams. This is the default behavior.
        """,
    )
    stream.add_argument(
        "--retry-max",
        metavar="COUNT",
        type=num(int, ge=0),
        help="""
            Stop fetching the list of available streams after `COUNT` retry attempt(s).

            A value of `0` makes Streamlink fetch streams indefinitely if --retry-streams is set to a non-zero value.
            If --retry-streams is unset, then the default delay between fetching available streams is 1 second.
        """,
    )
    stream.add_argument(
        "--retry-open",
        metavar="ATTEMPTS",
        type=num(int, ge=1),
        default=1,
        help="""
            After a successful fetch, try `ATTEMPTS` time(s) to open the stream until giving up.

            Default is 1.
        """,
    )
    stream.add_argument(
        "--stream-types",
        "--stream-priority",
        metavar="TYPES",
        type=comma_list,
        help="""
            A comma-delimited list of stream types to allow.

            The order will be used to separate streams when there are multiple
            streams with the same name but different stream types. Any stream type
            not listed will be omitted from the available streams list.  An `*` (asterisk) can
            be used as a wildcard to match any other type of stream, e.g. dash.

            Default is "hls,http,*".
        """,
    )
    stream.add_argument(
        "--stream-sorting-excludes",
        metavar="STREAMS",
        type=comma_list,
        help="""
            Fine-tune the `best` and `worst` stream name synonyms by excluding unwanted streams.

            If all of the available streams get excluded, `best` and `worst` will become
            inaccessible and new special stream synonyms `best-unfiltered` and `worst-unfiltered`
            can be used as a fallback selection method.

            The filter-expression's format is:

              [operator]<value>

            Valid operators are `>`, `>=`, `<` and `<=`. If no operator is specified then
            equality is tested.

            For example this will exclude streams ranked higher than "480p":

              --stream-sorting-excludes ">480p"

            Multiple filters can be used by separating each expression with a comma.

            For example this will exclude streams from two quality types:

              --stream-sorting-excludes ">480p,>medium"

        """,
    )

    transport = parser.add_argument_group("Stream transport options")
    transport_hls = parser.add_argument_group("HLS options", parent=transport)
    transport_dash = parser.add_argument_group("DASH options", parent=transport)
    transport_ffmpeg = parser.add_argument_group("FFmpeg options", parent=transport)

    transport.add_argument(
        "--ringbuffer-size",
        metavar="SIZE",
        type=filesize,
        help="""
            The maximum size of the ringbuffer.

            Mebibytes or kibibytes (base 2) can be specified via the M or K suffix respectively.

            The ringbuffer is used as a temporary storage between the stream and the player.
            This allows Streamlink to download the stream faster than the player which reads the data from the ringbuffer.

            The smaller the size of the ringbuffer, the higher the chance of the player buffering if the download speed
            decreases, and the higher the size, the more data can be use as a storage to recover from volatile download speeds.

            Most players have their own additional cache and will read the ringbuffer's content as soon as data is available.
            If the player stops reading data while playback is paused, Streamlink will continue to download the stream in the
            background as long as the ringbuffer doesn't get full.

            Default is "16M".
        """,
    )
    transport.add_argument(
        "--stream-segment-attempts",
        type=num(int, ge=1),
        metavar="ATTEMPTS",
        help="""
            The number of download attempts of each stream segment before giving up.

            This applies to all different kinds of segmented stream types, such as DASH, HLS, etc.

            Default is 3.
        """,
    )
    transport.add_argument(
        "--stream-segment-threads",
        type=num(int, ge=1, le=10),
        metavar="THREADS",
        help="""
            The size of the thread pool used to download segments. Minimum value is `1` and maximum is `10`.

            This applies to all different kinds of segmented stream types, such as DASH, HLS, etc.

            Default is 1.
        """,
    )
    transport.add_argument(
        "--stream-segment-timeout",
        type=num(float, gt=0),
        metavar="TIMEOUT",
        help="""
            The maximum time to wait for each segment to start downloading.

            This applies to all different kinds of segmented stream types, such as DASH, HLS, etc.

            Default is 10.0.
        """,
    )
    transport.add_argument(
        "--stream-segmented-duration",
        type=hours_minutes_seconds_float,
        metavar="[[XX:]XX:]XX[.XX] | [XXh][XXm][XX[.XX]s]",
        help="""
            Limit the output duration of segmented streams, like HLS and DASH.
            The actual duration may be slightly longer, as it is rounded to the nearest segment.

            Set to 0 to disable.

            Default is 0.
        """,
    )
    transport.add_argument(
        "--stream-segmented-queue-deadline",
        metavar="FACTOR",
        type=num(float, ge=0.0),
        help="""
            A multiplication factor of the time frame in which new segments must be queued in order to prevent playback issues
            due to lack of video/audio data. If this segment-queue-deadline has not been met, the stream will be stopped early.

            The intention of this segment-queue-deadline is to be able to stop early when the end of a stream is not announced
            by the server, so Streamlink doesn't have to wait until a buffer read-timeout occurs. See --stream-timeout.

            The base time this multiplication factor is applied to depends on the specific
            stream implementation and the respective values returned by the streaming server.
            This deadline check is done after trying to fetch new data.

            Set to ``0`` to disable.

            Default is 3.0.

            By default, wait three times as long for new segments to be made available than the server's advertised time frame.
        """,
    )
    transport.add_argument(
        "--stream-timeout",
        type=num(float, gt=0),
        metavar="TIMEOUT",
        help="""
            The maximum time to wait for an unfiltered stream to continue outputting data.

            This applies to all different kinds of stream types, such as DASH, HLS, HTTP, etc.

            Default is 60.0.
        """,
    )
    transport.add_argument(
        "--stream-passthrough-encrypted",
        action="store_true",
        default=None,
        help="""
            Pass through data from encrypted streams without decryption or encryption checks.

            This applies to DASH and HLS streams, and will likely result in garbage output.
        """,
    )
    transport.add_argument(
        "--mux-subtitles",
        action="store_true",
        default=None,
        help="""
            Automatically mux available subtitles into the output stream.

            Needs to be supported by the used plugin.
        """,
    )

    transport_hls.add_argument(
        "--hls-live-edge",
        type=num(int, ge=1),
        metavar="SEGMENTS",
        help="""
            Number of segments from the live stream's current live position to begin streaming.
            The size or length of each segment is determined by the streaming provider.

            Lower values will decrease the latency, but will also increase the chance of buffering, as there is less time for
            Streamlink to download segments and write their data to the output buffer. The number of parallel segment downloads
            can be set with --stream-segment-threads and the HLS playlist reload time to fetch and queue new segments can be
            overridden with --hls-playlist-reload-time.

            Default is 3.

            Note: During live playback, the caching/buffering settings of the used player will add additional latency.
            To adjust this, please refer to the player's own documentation for the required configuration.
            Player parameters can be set via --player-args.
        """,
    )
    transport_hls.add_argument(
        "--hls-segment-stream-data",
        action="store_true",
        default=None,
        help="""
            Immediately write segment data into output buffer while downloading.
        """,
    )
    transport_hls.add_argument(
        "--hls-playlist-reload-attempts",
        type=num(int, ge=1),
        metavar="ATTEMPTS",
        help="""
            The maximum number of attempts when reloading the HLS playlist before giving up.

            Default is 3.
        """,
    )
    transport_hls.add_argument(
        "--hls-playlist-reload-time",
        metavar="TIME",
        help="""
            Set a custom HLS playlist reload time value, either in seconds or by using one of the following keywords:

            - segment: The duration of the last segment in the current playlist
            - live-edge: The sum of segment durations of the live edge value minus one
            - default: The playlist's target duration metadata

            Default is default.
        """,
    )
    transport_hls.add_argument(
        "--hls-segment-queue-threshold",
        metavar="FACTOR",
        type=num(float, ge=0.0),
        help="""
            Deprecated in favor of --stream-segmented-queue-deadline.
        """,
    )
    transport_hls.add_argument(
        "--hls-segment-ignore-names",
        metavar="NAMES",
        type=comma_list,
        help="""
            A comma-delimited list of segment names that will get filtered out.

            Example: --hls-segment-ignore-names 000,001,002

            This will ignore every segment that ends with 000.ts, 001.ts and 002.ts

            Default is None.
        """,
    )
    transport_hls.add_argument(
        "--hls-segment-key-uri",
        metavar="URI",
        type=str,
        help="""
            Override the segment encryption key URIs for encrypted streams.

            The value can be templated using the following variables, which will be
            replaced with their respective part from the source segment URI:

              {url} {scheme} {netloc} {path} {query}

            Examples:

              --hls-segment-key-uri "https://example.com/hls/encryption_key"
              --hls-segment-key-uri "{scheme}://1.2.3.4{path}{query}"
              --hls-segment-key-uri "{scheme}://{netloc}/custom/path/to/key"

            Default is None.
        """,
    )
    transport_hls.add_argument(
        "--hls-audio-select",
        type=comma_list,
        metavar="CODE",
        help="""
            Select one or more specific audio sources by language code or name.
            Can be set to `*` (asterisk) to include all audio sources.

            Examples:

              --hls-audio-select "English,German"
              --hls-audio-select "en,de"
              --hls-audio-select "*"

            Note: This is only useful in special circumstances where the regular
            locale option fails, such as when multiple sources of the same language exist.
        """,
    )
    transport_hls.add_argument(
        "--hls-start-offset",
        type=hours_minutes_seconds_float,
        metavar="[[XX:]XX:]XX[.XX] | [XXh][XXm][XX[.XX]s]",
        help="""
            The amount of time to skip from the beginning of the stream.
            For live streams, this is a negative offset from the end of the stream (rewind).

            Default is 0.
        """,
    )
    transport_hls.add_argument(
        "--hls-duration",
        type=hours_minutes_seconds_float,
        metavar="[[XX:]XX:]XX[.XX] | [XXh][XXm][XX[.XX]s]",
        help="""
            Deprecated in favor of --stream-segmented-duration.
        """,
    )
    transport_hls.add_argument(
        "--hls-live-restart",
        action="store_true",
        default=None,
        help="""
            Skip to the beginning of a live stream, or as far back as possible.
        """,
    )

    transport_dash.add_argument(
        "--dash-manifest-reload-attempts",
        type=num(int, ge=1),
        metavar="ATTEMPTS",
        help="""
            The maximum number of attempts when reloading the DASH manifest before giving up.

            Default is 3.
        """,
    )

    transport_ffmpeg.add_argument(
        "--ffmpeg-ffmpeg",
        metavar="FILENAME",
        help="""
            Set the location of the FFmpeg executable if it can't be resolved
            from the paths of the system's `PATH` environment variable.

            FFmpeg is required to access or mux separate video and audio streams,
            e.g. in DASH streams or HLS streams with multiple sources.

            Example: --ffmpeg-ffmpeg "/usr/local/bin/ffmpeg"
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-no-validation",
        action="store_true",
        default=None,
        help="""
            Disable FFmpeg validation and version logging.
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-verbose",
        action="store_true",
        default=None,
        help="""
            Write FFmpeg's stderr output to Streamlink's stderr output.
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-verbose-path",
        type=str,
        metavar="PATH",
        help="""
            Write FFmpeg's stderr output to PATH.
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-loglevel",
        type=str,
        metavar="LOGLEVEL",
        help="""
            Change FFmpeg's `-loglevel` value to `LOGLEVEL`.

            Unless --ffmpeg-verbose or --ffmpeg-verbose-path is set, changing the log level won't have any effect.

            Default is "info".
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-fout",
        type=str,
        metavar="OUTFORMAT",
        help="""
            Set the output format to `OUTFORMAT`. This only applies to streams which require muxing.

            Default is "matroska".

            Example: --ffmpeg-fout "mpegts"
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-video-transcode",
        metavar="CODEC",
        help="""
            Transcode the video to `CODEC`. This only applies to streams which require muxing.

            Default is "copy".

            Example: --ffmpeg-video-transcode "h264"
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-audio-transcode",
        metavar="CODEC",
        help="""
            Transcode the audio to `CODEC`. This only applies to streams which require muxing.

            Default is "copy".

            Example: --ffmpeg-audio-transcode "aac"
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-copyts",
        action="store_true",
        default=None,
        help="""
            Set the `-copyts` FFmpeg option, so input timestamps won't be processed
            and the initial start time offset value be kept.
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-start-at-zero",
        action="store_true",
        default=None,
        help="""
            Enable the `-start_at_zero` FFmpeg option when using --ffmpeg-copyts.
        """,
    )
    transport_ffmpeg.add_argument(
        "--ffmpeg-validation-timeout",
        type=float,
        metavar="SECONDS",
        help="""
            Timeout in seconds for FFmpeg version validation.

            Default is 4.0.

            Increase this on low-power systems if FFmpeg startup is slow.
        """,
    )

    http = parser.add_argument_group("HTTP options")
    http.add_argument(
        "--http-proxy",
        metavar="HTTP_PROXY",
        help="""
            An HTTP proxy to use for all HTTP and HTTPS requests, including WebSocket connections.

            Example: --http-proxy "http://hostname:port/"
        """,
    )
    http.add_argument("--https-proxy", help=argparse.SUPPRESS)
    http.add_argument(
        "--http-cookie",
        metavar="KEY=VALUE",
        type=keyvalue,
        action="append",
        help="""
            A cookie to add to each HTTP request.

            Can be repeated to add multiple cookies.
        """,
    )
    http.add_argument(
        "--http-cookies-file",
        metavar="PATH",
        action="append",
        help="""
            A path to a cookies file whose cookie data will be added to HTTP requests.

            Can be repeated to add multiple cookie files.

            The file format must adhere to the Netscape HTTP Cookie File format (MozillaCookieJar, curl, etc.),
            also known as cookies.txt. Be aware that this format loses information about RFC 2965 cookies,
            and also about newer or non-standard cookie-attributes such as port.

            Unlike --http-cookie, which sends data with every request, this option enables granular control
            by respecting domain, path, and other cookie attributes.
        """,
    )
    http.add_argument(
        "--http-header",
        metavar="KEY=VALUE",
        type=keyvalue,
        action="append",
        help="""
            A header to add to each HTTP request.

            Can be repeated to add multiple headers.
        """,
    )
    http.add_argument(
        "--http-query-param",
        metavar="KEY=VALUE",
        type=keyvalue,
        action="append",
        help="""
            A query parameter to add to each HTTP request.

            Can be repeated to add multiple query parameters.
        """,
    )
    http.add_argument(
        "--http-ignore-env",
        action="store_false",
        default=None,
        help="""
            Ignore HTTP settings set in the environment, such as environment variables (`HTTP_PROXY`, etc)
            or `~/.netrc` authentication.
        """,
    )
    http.add_argument(
        "--http-no-ssl-verify",
        action="store_false",
        default=None,
        help="""
            Don't attempt to verify TLS/SSL certificates.

            Use with caution, as it has TLS/SSL security implications.
        """,
    )
    http.add_argument(
        "--http-disable-dh",
        action="store_true",
        default=None,
        help="""
            Disable Diffie Hellman key exchange.

            Use with caution, as it has TLS/SSL security implications.
        """,
    )
    http.add_argument(
        "--http-ssl-cert",
        metavar="PEM_FILENAME",
        help="""
            SSL certificate to use: a .pem file.
        """,
    )
    http.add_argument(
        "--http-ssl-cert-crt-key",
        metavar=("CRT_FILENAME", "KEY_FILENAME"),
        nargs=2,
        help="""
            SSL certificate to use: a .crt and a .key file.
        """,
    )
    http.add_argument(
        "--http-timeout",
        metavar="TIMEOUT",
        type=num(float, gt=0),
        help="""
            Set the general timeout value used by all HTTP requests except the ones covered by other options.

            Default is 20.0.
        """,
    )

    webbrowser = parser.add_argument_group("Web browser options")
    webbrowser.add_argument(
        "--webbrowser",
        type=boolean,
        metavar="{yes,true,1,on,no,false,0,off}",
        default=None,
        help="""
            Enable or disable support for Streamlink's webbrowser API.

            Streamlink's webbrowser API allows plugins which implement it to launch a web browser and extract data from websites
            which they otherwise couldn't do via the regular HTTP session in Python due to specific JavaScript restrictions.

            The web browser is run isolated and in a clean environment without access to regular user data.

            Streamlink currently only supports Chromium-based web browsers using the Chrome Devtools Protocol (CDP).
            This includes Chromium itself, Google Chrome, Microsoft Edge, Brave, Vivaldi, and others, but full support for
            third party Chromium forks is not guaranteed. Please try Chromium or Google Chrome when encountering any issues.

            Default is true.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-executable",
        metavar="PATH",
        help="""
            Path to the web browser's executable.

            By default, it is looked up automatically according to the rules of the used webbrowser API implementation.
            This usually involves a list of known executable names and fallback paths on all supported operating systems.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-timeout",
        metavar="TIME",
        type=num(float, gt=0),
        help="""
            The maximum amount of time which the web browser can take to launch and execute.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-cdp-host",
        metavar="HOST",
        help="""
            Host for the web browser's inter-process communication interface (CDP specific).

            Default is 127.0.0.1.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-cdp-port",
        metavar="PORT",
        type=num(int, ge=0, le=65535),
        help="""
            Port for the web browser's inter-process communication interface (CDP specific).

            Tries to find a free port by default.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-cdp-timeout",
        metavar="TIME",
        type=num(float, gt=0),
        help="""
            The maximum amount of time for waiting on a single CDP command response.
        """,
    )
    webbrowser.add_argument(
        "--webbrowser-headless",
        type=boolean,
        metavar="{yes,true,1,on,no,false,0,off}",
        default=None,
        help="""
            Whether to launch the web browser in headless mode or not.
            When enabled, it stays completely hidden and doesn't require a desktop environment to run.

            Please be aware that headless mode might be blocked by websites which implement bot detections.

            Default is false.
        """,
    )

    return parser


# The order of arguments determines if options get overridden by `Streamlink.set_option()`
# NOTE: arguments with `action=store_{true,false}` must set `default=None`
_ARGUMENT_TO_SESSIONOPTION: list[tuple[str, str, Callable[[Any], Any] | type | None]] = [
    # generic arguments
    ("no_plugin_cache", "no-plugin-cache", None),
    ("locale", "locale", None),
    # network arguments
    ("interface", "interface", None),
    ("ipv4", "ipv4", None),
    ("ipv6", "ipv6", None),
    # HTTP session arguments
    ("https_proxy", "https-proxy", None),
    ("http_proxy", "http-proxy", None),
    ("http_cookies_file", "http-cookies-files", None),
    ("http_cookie", "http-cookies", dict),
    ("http_header", "http-headers", dict),
    ("http_query_param", "http-query-params", dict),
    ("http_ignore_env", "http-trust-env", None),
    ("http_no_ssl_verify", "http-ssl-verify", None),
    ("http_disable_dh", "http-disable-dh", None),
    ("http_ssl_cert", "http-ssl-cert", None),
    ("http_ssl_cert_crt_key", "http-ssl-cert", tuple),
    ("http_timeout", "http-timeout", None),
    # stream transport arguments
    ("hls_duration", "hls-duration", None),  # deprecated options must come first
    ("hls_segment_queue_threshold", "hls-segment-queue-threshold", None),  # deprecated options must come first
    ("ringbuffer_size", "ringbuffer-size", None),
    ("mux_subtitles", "mux-subtitles", None),
    ("stream_segment_attempts", "stream-segment-attempts", None),
    ("stream_segment_threads", "stream-segment-threads", None),
    ("stream_segment_timeout", "stream-segment-timeout", None),
    ("stream_segmented_duration", "stream-segmented-duration", None),
    ("stream_segmented_queue_deadline", "stream-segmented-queue-deadline", None),
    ("stream_timeout", "stream-timeout", None),
    ("stream_passthrough_encrypted", "stream-passthrough-encrypted", None),
    ("hls_live_edge", "hls-live-edge", None),
    ("hls_live_restart", "hls-live-restart", None),
    ("hls_start_offset", "hls-start-offset", None),
    ("hls_playlist_reload_attempts", "hls-playlist-reload-attempts", None),
    ("hls_playlist_reload_time", "hls-playlist-reload-time", None),
    ("hls_segment_stream_data", "hls-segment-stream-data", None),
    ("hls_segment_ignore_names", "hls-segment-ignore-names", None),
    ("hls_segment_key_uri", "hls-segment-key-uri", None),
    ("hls_audio_select", "hls-audio-select", None),
    ("dash_manifest_reload_attempts", "dash-manifest-reload-attempts", None),
    ("ffmpeg_ffmpeg", "ffmpeg-ffmpeg", None),
    ("ffmpeg_no_validation", "ffmpeg-no-validation", None),
    ("ffmpeg_verbose", "ffmpeg-verbose", None),
    ("ffmpeg_verbose_path", "ffmpeg-verbose-path", None),
    ("ffmpeg_loglevel", "ffmpeg-loglevel", None),
    ("ffmpeg_fout", "ffmpeg-fout", None),
    ("ffmpeg_video_transcode", "ffmpeg-video-transcode", None),
    ("ffmpeg_audio_transcode", "ffmpeg-audio-transcode", None),
    ("ffmpeg_copyts", "ffmpeg-copyts", None),
    ("ffmpeg_start_at_zero", "ffmpeg-start-at-zero", None),
    # web browser arguments
    ("webbrowser", "webbrowser", None),
    ("webbrowser_executable", "webbrowser-executable", None),
    ("webbrowser_timeout", "webbrowser-timeout", None),
    ("webbrowser_cdp_host", "webbrowser-cdp-host", None),
    ("webbrowser_cdp_port", "webbrowser-cdp-port", None),
    ("webbrowser_cdp_timeout", "webbrowser-cdp-timeout", None),
    ("webbrowser_headless", "webbrowser-headless", None),
    ("ffmpeg_validation_timeout", "ffmpeg-validation-timeout", None),
]


def setup_session_options(session: Streamlink, args: argparse.Namespace):
    for arg, option, mapper in _ARGUMENT_TO_SESSIONOPTION:
        value = getattr(args, arg)
        if value is not None:
            if mapper is not None:
                value = mapper(value)
            session.set_option(option, value)


def setup_plugin_args(session: Streamlink, parser: ArgumentParser):
    """Adds plugin argument data to the argument parser."""

    plugin_args = parser.add_argument_group("Plugin-specific options")
    for pname, arguments in session.plugins.iter_arguments():
        group = parser.add_argument_group(pname.capitalize(), parent=plugin_args)

        for parg in arguments:
            group.add_argument(parg.argument_name(pname), **parg.options)


def setup_plugin_options(
    session: Streamlink,
    args: argparse.Namespace,
    pluginname: str,
    pluginclass: type[Plugin],
) -> Options:
    """Initializes plugin options from argument values."""

    if not pluginclass.arguments:
        return Options()

    user_input_requester: UserInputRequester | None = session.get_option("user-input-requester")
    if not user_input_requester:
        raise RuntimeError("The Streamlink session is missing a UserInputRequester")

    defaults = {}
    values = {}
    required = {}

    for parg in pluginclass.arguments:
        value = getattr(args, parg.namespace_dest(pluginname))
        values[parg.dest] = value
        defaults[parg.dest] = parg.default

        if parg.help == argparse.SUPPRESS:
            if value != parg.default:
                warnings.warn(
                    f"The {parg.argument_name(pluginname)} plugin argument has been disabled and will be removed in the future",
                    StreamlinkDeprecationWarning,
                    stacklevel=1,
                )
            continue

        if parg.required:
            required[parg.name] = parg
        # if the value is set, check to see if any of the required arguments are not set
        if parg.required or value:
            try:
                for rparg in pluginclass.arguments.requires(parg.name):
                    required[rparg.name] = rparg
            except RuntimeError:  # pragma: no cover
                log.error(f"{pluginname} plugin has a configuration error and the arguments cannot be parsed")
                break

    for req in required.values():
        if not values.get(req.dest):
            prompt = f"{req.prompt or f'Enter {pluginname} {req.name}'}"
            try:
                if req.sensitive:
                    value = user_input_requester.ask_password(prompt)
                else:
                    value = user_input_requester.ask(prompt)
            except OSError as err:
                raise StreamlinkCLIError from err
            values[req.dest] = value

    options = Options(defaults)
    options.update(values)

    return options


__all__ = ["ArgumentParser", "build_parser", "setup_session_options", "setup_plugin_args", "setup_plugin_options"]



from __future__ import annotations

import sys
from ctypes import POINTER, Structure, byref
from ctypes.wintypes import (
    BOOL,
    DWORD,
    HANDLE,
    SHORT,
    SMALL_RECT,
    WCHAR,
    WORD,
)
from typing import TYPE_CHECKING, ClassVar


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from ctypes import CDLL
    from io import TextIOWrapper


# https://learn.microsoft.com/en-us/windows/console/coord-str
class COORD(Structure):
    _fields_: ClassVar = [
        ("X", SHORT),
        ("Y", SHORT),
    ]


# https://learn.microsoft.com/en-us/windows/console/console-screen-buffer-info-str
# noinspection PyPep8Naming
class CONSOLE_SCREEN_BUFFER_INFO(Structure):
    _fields_: ClassVar = [
        ("dwSize", COORD),
        ("dwCursorPosition", COORD),
        ("wAttributes", WORD),
        ("srWindow", SMALL_RECT),
        ("dwMaximumWindowSize", COORD),
    ]


class _WinApiCall:
    argtypes: ClassVar[Sequence]
    restype: ClassVar
    method: Callable

    def __init__(self, dll: CDLL):
        self._dll = dll
        method = getattr(dll, self.__class__.__name__)
        method.argtypes = self.argtypes
        method.restype = self.restype
        self.__call__ = self.method = method

    def _call_success(self, *args):
        if not self.method(*args):
            # https://learn.microsoft.com/en-us/windows/win32/api/errhandlingapi/nf-errhandlingapi-getlasterror
            # https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes#system-error-codes
            last_error = self._dll.GetLastError()

            raise OSError(f"Error while calling kernel32.{self.__class__.__name__} ({last_error=:#x})")


class GetStdHandle(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/getstdhandle
    """

    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12

    argtypes: ClassVar = [DWORD]
    restype: ClassVar = HANDLE

    def __call__(self, handle: TextIOWrapper | None) -> HANDLE:
        if handle is sys.stderr:
            std_handle = self.STD_ERROR_HANDLE
        else:
            std_handle = self.STD_OUTPUT_HANDLE

        return self.method(std_handle)


class GetConsoleMode(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/getconsolemode
    """

    ENABLE_VIRTUAL_TERMINAL_PROCESSING = 4

    argtypes: ClassVar = [HANDLE, POINTER(DWORD)]
    restype: ClassVar = BOOL

    def __call__(self, console_output: HANDLE) -> int:
        mode = DWORD()
        self._call_success(console_output, mode)

        return mode.value

    def supports_virtual_terminal_processing(self, console_output: HANDLE) -> bool:
        try:
            console_mode = self(console_output)
        except OSError:
            return False

        return console_mode & self.ENABLE_VIRTUAL_TERMINAL_PROCESSING > 0


class GetConsoleScreenBufferInfo(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/getconsolescreenbufferinfo
    """

    argtypes: ClassVar = [HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFO)]
    restype: ClassVar = BOOL

    def __call__(self, console_output: HANDLE) -> CONSOLE_SCREEN_BUFFER_INFO:
        console_screen_buffer_info = CONSOLE_SCREEN_BUFFER_INFO()
        self._call_success(console_output, byref(console_screen_buffer_info))

        return console_screen_buffer_info


class SetConsoleCursorPosition(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/setconsolecursorposition
    """

    argtypes: ClassVar = [HANDLE, COORD]
    restype: ClassVar = BOOL

    def __call__(self, console_output: HANDLE, cursor_position: COORD) -> None:
        self._call_success(console_output, cursor_position)


class FillConsoleOutputAttribute(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/fillconsoleoutputattribute
    """

    argtypes: ClassVar = [HANDLE, WORD, DWORD, COORD, POINTER(DWORD)]
    restype: ClassVar = BOOL

    def __call__(self, console_output: HANDLE, attribute: int, length: int, write_coord: COORD) -> int:
        attrs = WORD(attribute)
        size = DWORD(length)
        number_of_attrs_written = DWORD()
        self._call_success(console_output, attrs, size, write_coord, byref(number_of_attrs_written))

        return number_of_attrs_written.value


class FillConsoleOutputCharacterW(_WinApiCall):
    """
    https://learn.microsoft.com/en-us/windows/console/fillconsoleoutputcharacter
    """

    argtypes: ClassVar = [HANDLE, WCHAR, DWORD, COORD, POINTER(DWORD)]
    restype: ClassVar = BOOL

    def __call__(self, console_output: HANDLE, character: str, length: int, write_coord: COORD) -> int:
        char = WCHAR(character)
        size = DWORD(length)
        number_of_chars_written = DWORD()
        self._call_success(console_output, char, size, write_coord, byref(number_of_chars_written))

        return number_of_chars_written.value


class WindowsConsole:
    get_std_handle: ClassVar[GetStdHandle]
    get_console_mode: ClassVar[GetConsoleMode]
    get_console_screen_buffer_info: ClassVar[GetConsoleScreenBufferInfo]
    set_console_cursor_position: ClassVar[SetConsoleCursorPosition]
    fill_console_output_attribute: ClassVar[FillConsoleOutputAttribute]
    fill_console_output_character_w: ClassVar[FillConsoleOutputCharacterW]

    def __new__(cls, *args, **kwargs):
        try:
            from ctypes import windll  # type: ignore[import]  # noqa: PLC0415
        except ImportError:  # pragma: no cover
            return None

        kernel32 = windll.kernel32
        cls.get_std_handle = GetStdHandle(kernel32)
        cls.get_console_mode = GetConsoleMode(kernel32)
        cls.get_console_screen_buffer_info = GetConsoleScreenBufferInfo(kernel32)
        cls.set_console_cursor_position = SetConsoleCursorPosition(kernel32)
        cls.fill_console_output_attribute = FillConsoleOutputAttribute(kernel32)
        cls.fill_console_output_character_w = FillConsoleOutputCharacterW(kernel32)

        return super().__new__(cls)

    def __init__(self, handle: TextIOWrapper | None = None):
        self.handle = self.get_std_handle(handle)

    def supports_virtual_terminal_processing(self):
        return self.get_console_mode.supports_virtual_terminal_processing(self.handle)

    def clear_line(self):
        info = self.get_console_screen_buffer_info(self.handle)

        def_attrs = info.wAttributes
        length = info.dwSize.X
        cursor_position = COORD(X=0, Y=info.dwCursorPosition.Y)
        self.fill_console_output_character_w(self.handle, " ", length, cursor_position)
        self.fill_console_output_attribute(self.handle, def_attrs, length, cursor_position)
        self.set_console_cursor_position(self.handle, cursor_position)



# ruff: noqa: F401
import hashlib


# re-export pycryptodome / pycryptodomex stuff in a single place
# so packagers don't have to maintain dozens of patches
try:
    # pycryptodome (drop-in replacement for the old PyCrypto library)
    from Crypto.Cipher import AES, PKCS1_v1_5
    from Crypto.Hash import SHA256
    from Crypto.PublicKey import RSA
    from Crypto.Util.Padding import pad, unpad
except ImportError:  # pragma: no cover
    # pycryptodomex (independent of the old PyCrypto library)
    from Cryptodome.Cipher import AES, PKCS1_v1_5  # type: ignore
    from Cryptodome.Hash import SHA256  # type: ignore
    from Cryptodome.PublicKey import RSA  # type: ignore
    from Cryptodome.Util.Padding import pad, unpad  # type: ignore


def evp_bytestokey(password, salt, key_len, iv_len):
    """
    Python implementation of OpenSSL's EVP_BytesToKey()
    :param password: or passphrase
    :param salt: 8 byte salt
    :param key_len: length of key in bytes
    :param iv_len:  length of IV in bytes
    :return: (key, iv)
    """
    d = d_i = b""
    while len(d) < key_len + iv_len:
        d_i = hashlib.md5(d_i + password + salt).digest()
        d += d_i
    return d[:key_len], d[key_len : key_len + iv_len]


def decrypt_openssl(data, passphrase, key_length=32):
    if data.startswith(b"Salted__"):
        salt = data[len(b"Salted__") : AES.block_size]
        key, iv = evp_bytestokey(passphrase, salt, key_length, AES.block_size)
        d = AES.new(key, AES.MODE_CBC, iv)
        out = d.decrypt(data[AES.block_size :])
        return unpad_pkcs5(out)


def unpad_pkcs5(padded):
    return padded[: -padded[-1]]



"""
$description Esports tournaments run by BlastTV, based in Denmark.
$url blast.tv
$type live, vod
$metadata id
$metadata title
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    name="live",
    pattern=re.compile(r"^https?://(?:www\.)?blast\.tv(?:/?$|/live(?:/(?P<channel>[a-z0-9_-]+))?)"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"^https?://(?:www\.)?blast\.tv/(?P<game>[^/]+)/tournaments/[^/]+/match/(?P<shortid>\w+)/"),
)
class BlastTv(Plugin):
    _URL_API_LIVE = "https://api.blast.tv/v1/broadcasts/live"
    _URL_API_MATCHES = "https://api.blast.tv/v2/games/{game}/matches/{shortid}"
    _URL_API_REWATCH = "https://api.blast.tv/v1/videos/rewatch/{id}"

    def _get_live(self):
        channel = self.match["channel"]

        live_channels = self.session.http.get(
            self._URL_API_LIVE,
            schema=validate.Schema(
                validate.parse_json(),
                [
                    {
                        "id": str,
                        "slug": str,
                        "priority": int,
                        "title": str,
                        "videoSrc": validate.any("", validate.url(path=validate.endswith(".m3u8"))),
                        "videoAlternativeSrc": validate.any("", validate.url(scheme="http")),
                    },
                ],
            ),
        )

        for live_channel in sorted(live_channels, key=lambda x: x["priority"]):
            if channel and channel != live_channel["slug"]:
                continue

            if live_channel["videoSrc"]:
                self.id = live_channel["id"]
                self.title = live_channel["title"]
                return HLSStream.parse_variant_playlist(self.session, live_channel["videoSrc"])

            if live_channel["videoAlternativeSrc"]:
                return self.session.streams(live_channel["videoAlternativeSrc"])

    def _get_vod(self):
        self.id, external_stream_url = self.session.http.get(
            self._URL_API_MATCHES.format(
                game=self.match["game"],
                shortid=self.match["shortid"],
            ),
            acceptable_status=(200, 404),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("id"): str,
                    validate.optional("metadata"): validate.all(
                        {
                            validate.optional("externalStreamUrl"): validate.any("", validate.url(scheme="http")),
                        },
                        validate.get("externalStreamUrl"),
                    ),
                },
                validate.union_get(
                    "id",
                    "metadata",
                ),
            ),
        )
        if not self.id:
            return

        if external_stream_url:
            return self.session.streams(external_stream_url)

        hls_url = self.session.http.get(
            self._URL_API_REWATCH.format(id=self.id),
            acceptable_status=(200, 404),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("src"): validate.url(path=validate.endswith(".m3u8")),
                },
                validate.get("src"),
            ),
        )
        if not hls_url:
            return

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams(self):
        if self.matches["live"]:
            return self._get_live()
        if self.matches["vod"]:
            return self._get_vod()


__plugin__ = BlastTv



"""
$description European live TV channels and video on-demand service. OTT service from Zattoo.
$url zattoo.com
$url iptv.glattvision.ch
$url www.saktv.ch
$url www.vtxtv.ch
$url mobiltv.quickline.com
$url www.quantum-tv.com
$url tvonline.ewe.de
$url nettv.netcologne.de
$url tvplus.m-net.de
$url player.waly.tv
$url www.1und1.tv
$url www.netplus.tv
$url www.bbv-tv.net
$url www.meinewelt.cc
$type live, vod
"""

import re
import uuid

from streamlink.cache import Cache
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(
        r"""
            https?://
            (?P<base_url>
                (?:
                    iptv\.glattvision|www\.(?:saktv|vtxtv)
                )\.ch
                |(?:
                    mobiltv\.quickline|www\.quantum-tv|zattoo
                )\.com
                |(?:
                    tvonline\.ewe|nettv\.netcologne|tvplus\.m-net
                )\.de
                |(?:
                    player\.waly|www\.(?:1und1|netplus)
                )\.tv
                |www\.bbv-tv\.net
                |www\.meinewelt\.cc
            )/
            (?:
                (?:
                    recording(?:s\?recording=|/)
                    |
                    (?:ondemand/)?watch/[^/\s]+/[^/]+/
                )(?P<recording_id>\d+)
                |
                (?:
                    (?:live/|watch/)|(?:channels(?:/\w+)?|guide)\?channel=
                )(?P<channel>[^/\s]+)
                |
                ondemand(?:\?video=|/watch/)(?P<vod_id>[^-]+)
            )
        """,
        re.VERBOSE,
    ),
)
@pluginargument(
    "email",
    requires=["password"],
    metavar="EMAIL",
    help="The email associated with your zattoo account, required to access any zattoo stream.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="A zattoo account password to use with --zattoo-email.",
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached zattoo credentials to initiate a new session and reauthenticate.",
)
@pluginargument(
    "stream-types",
    metavar="TYPES",
    type="comma_list_filter",
    type_kwargs={"acceptable": ["dash", "hls7"]},
    default=["dash"],
    help="""
        A comma-delimited list of stream types which should be used.

        The following types are allowed: dash, hls7

        Default is "dash".
    """,
)
class Zattoo(Plugin):
    TIME_CONTROL = 60 * 60 * 2
    TIME_SESSION = 60 * 60 * 24 * 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain = self.match.group("base_url")
        self._session_attributes = Cache(
            filename="plugin-cache.json",
            key_prefix=f"zattoo:attributes:{self.domain}",
        )
        self._uuid = self._session_attributes.get("uuid")
        self._authed = (
            self._session_attributes.get("power_guide_hash")
            and self._uuid
            and self.session.http.cookies.get("pzuid", domain=self.domain)
            and self.session.http.cookies.get("beaker.session.id", domain=self.domain)
        )
        self._session_control = self._session_attributes.get("session_control", False)
        self.base_url = f"https://{self.domain}"
        self.headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "X-Requested-With": "XMLHttpRequest",
            "Referer": self.base_url,
        }

    def _hello(self):
        log.debug("_hello ...")
        app_token = self.session.http.get(
            f"{self.base_url}/token.json",
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "success": bool,
                    "session_token": str,
                },
                validate.get("session_token"),
            ),
        )
        if self._uuid:
            uuid_data = self._uuid
        else:
            uuid_data = str(uuid.uuid4())
            self._session_attributes.set("uuid", uuid_data, expires=self.TIME_SESSION)

        params = {
            "app_version": "3.2120.1",
            "client_app_token": app_token,
            "format": "json",
            "lang": "en",
            "uuid": uuid_data,
        }
        res = self.session.http.post(
            f"{self.base_url}/zapi/v3/session/hello",
            headers=self.headers,
            data=params,
            schema=validate.Schema(
                validate.parse_json(),
                validate.any({"active": bool}, {"success": bool}),
            ),
        )
        if res.get("active") or res.get("success"):
            log.debug("Hello was successful.")
        else:
            log.debug("Hello failed.")

    def _login(self, email, password):
        log.debug("_login ...")
        data = self.session.http.post(
            f"{self.base_url}/zapi/v3/account/login",
            headers=self.headers,
            data={
                "login": email,
                "password": password,
                "remember": "true",
                "format": "json",
            },
            acceptable_status=(200, 400),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {"active": bool, "power_guide_hash": str},
                    {"success": bool},
                ),
            ),
        )

        if data.get("active"):
            log.debug("Login was successful.")
        else:
            log.debug("Login failed.")
            return

        self._authed = data["active"]
        self.save_cookies(default_expires=self.TIME_SESSION)
        self._session_attributes.set("power_guide_hash", data["power_guide_hash"], expires=self.TIME_SESSION)
        self._session_attributes.set("session_control", True, expires=self.TIME_CONTROL)

    def _watch(self):
        log.debug("_watch ...")
        channel = self.match.group("channel")
        vod_id = self.match.group("vod_id")
        recording_id = self.match.group("recording_id")

        params = {"https_watch_urls": True}
        if channel:
            watch_url = f"{self.base_url}/zapi/watch"
            params_cid = self._get_params_cid(channel)
            if not params_cid:
                return
            params.update(params_cid)
        elif vod_id:
            log.debug(f"Found vod_id: {vod_id}")
            watch_url = f"{self.base_url}/zapi/avod/videos/{vod_id}/watch"
        elif recording_id:
            log.debug(f"Found recording_id: {recording_id}")
            watch_url = f"{self.base_url}/zapi/watch/recording/{recording_id}"
        else:
            log.debug("Missing watch_url")
            return

        zattoo_stream_types = self.get_option("stream-types")
        for stream_type in zattoo_stream_types:
            params_stream_type = {"stream_type": stream_type}
            params.update(params_stream_type)

            data = self.session.http.post(
                watch_url,
                headers=self.headers,
                data=params,
                acceptable_status=(200, 402, 403, 404),
                schema=validate.Schema(
                    validate.parse_json(),
                    validate.any(
                        {
                            "success": validate.transform(bool),
                            "stream": {
                                "watch_urls": [
                                    {
                                        "url": validate.url(),
                                        validate.optional("maxrate"): int,
                                        validate.optional("audio_channel"): str,
                                    },
                                ],
                                validate.optional("quality"): str,
                            },
                        },
                        {
                            "success": validate.transform(bool),
                            "internal_code": int,
                            validate.optional("http_status"): int,
                        },
                    ),
                ),
            )

            if not data["success"]:
                if data["internal_code"] == 401:
                    log.error(f"invalid stream_type {stream_type}")
                elif data["internal_code"] == 421:
                    log.error("Unfortunately streaming is not permitted in this country or this channel does not exist.")
                elif data["internal_code"] == 422:
                    log.error("Paid subscription required for this channel.")
                    log.info("If paid subscription exist, use --zattoo-purge-credentials to start a new session.")
                else:
                    log.debug("unknown error %r", data)
                    log.debug("Force session reset for watch_url")
                    self.reset_session()
                continue

            log.debug(f"Found data for {stream_type}")
            if stream_type == "hls7":
                for url in data["stream"]["watch_urls"]:
                    yield from HLSStream.parse_variant_playlist(
                        self.session,
                        url["url"],
                        ffmpeg_options={"copyts": True},
                    ).items()
            elif stream_type == "dash":
                for url in data["stream"]["watch_urls"]:
                    yield from DASHStream.parse_manifest(self.session, url["url"]).items()

    def _get_params_cid(self, channel):
        log.debug(f"get channel ID for {channel}")
        try:
            res = self.session.http.get(
                f"{self.base_url}/zapi/v2/cached/channels/{self._session_attributes.get('power_guide_hash')}",
                headers=self.headers,
                params={"details": "False"},
            )
        except Exception:
            log.debug("Force session reset for _get_params_cid")
            self.reset_session()
            return False

        data = self.session.http.json(
            res,
            schema=validate.Schema(
                {
                    "success": validate.transform(bool),
                    "channel_groups": [
                        {
                            "channels": [
                                {
                                    "display_alias": str,
                                    "cid": str,
                                    "qualities": [
                                        {
                                            "title": str,
                                            "stream_types": validate.all(
                                                [str],
                                                validate.filter(
                                                    lambda n: not re.match(r"(.+_(?:fairplay|playready|widevine))", n),
                                                ),
                                            ),
                                            "level": str,
                                            "availability": str,
                                        },
                                    ],
                                },
                            ],
                        },
                    ],
                },
                validate.get("channel_groups"),
            ),
        )

        c_list = [channel for channel_group in data for channel in channel_group["channels"]]
        cid = None
        zattoo_list = []
        for c in c_list:
            zattoo_list.append(c["display_alias"])
            if c["display_alias"] == channel:
                cid = c["cid"]
                log.debug("%r", c)

        log.trace("Available zattoo channels in this country: %r", sorted(zattoo_list))

        if not cid:
            cid = channel

        log.debug(f"CHANNEL ID: {cid}")
        return {"cid": cid}

    def reset_session(self):
        self._session_attributes.set("power_guide_hash", None, expires=0)
        self._session_attributes.set("uuid", None, expires=0)
        self.clear_cookies()
        self._authed = False

    def _get_streams(self):
        email = self.get_option("email")
        password = self.get_option("password")

        if self.options.get("purge_credentials"):
            self.reset_session()
            log.info("All credentials were successfully removed.")
        elif self._authed and not self._session_control:
            # check every two hours, if the session is actually valid
            log.debug(f"Session control for {self.domain}")
            active = self.session.http.get(
                f"{self.base_url}/zapi/v3/session",
                schema=validate.Schema(validate.parse_json(), {"active": bool}, validate.get("active")),
            )
            if active:
                self._session_attributes.set("session_control", True, expires=self.TIME_CONTROL)
                log.debug("User is logged in")
            else:
                log.debug("User is not logged in")
                self._authed = False

        if not self._authed and (not email and not password):
            log.error("A login for Zattoo is required, use --zattoo-email EMAIL --zattoo-password PASSWORD to set them")
            return

        if not self._authed:
            self._hello()
            self._login(email, password)

        if self._authed:
            return self._watch()


__plugin__ = Zattoo



"""
$description Japanese live TV streaming website with multiple channels including news, sports, entertainment and anime.
$url abema.tv
$type live, vod
$region Japan
"""

import hashlib
import hmac
import re
import struct
import time
import uuid
from base64 import urlsafe_b64encode
from binascii import unhexlify

from requests import Response
from requests.adapters import BaseAdapter

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter
from streamlink.utils.crypto import AES
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


class AbemaTVHLSStreamWriter(HLSStreamWriter):
    def should_filter_segment(self, segment):
        return "/tsad/" in segment.uri or super().should_filter_segment(segment)


class AbemaTVHLSStreamReader(HLSStreamReader):
    __writer__ = AbemaTVHLSStreamWriter


class AbemaTVHLSStream(HLSStream):
    __reader__ = AbemaTVHLSStreamReader


class AbemaTVLicenseAdapter(BaseAdapter):
    """
    Handling abematv-license:// protocol to get real video key_data.
    """

    STRTABLE = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"

    HKEY = b"3AF0298C219469522A313570E8583005A642E73EDD58E3EA2FB7339D3DF1597E"

    _MEDIATOKEN_API = "https://api.abema.io/v1/media/token"

    _LICENSE_API = "https://license.abema.io/abematv-hls"

    _MEDIATOKEN_SCHEMA = validate.Schema({"token": str})

    _LICENSE_SCHEMA = validate.Schema({"k": str, "cid": str})

    def __init__(self, session, deviceid, usertoken):
        self._session = session
        self.deviceid = deviceid
        self.usertoken = usertoken
        super().__init__()

    def _get_videokey_from_ticket(self, ticket):
        params = {
            "osName": "android",
            "osVersion": "6.0.1",
            "osLang": "ja_JP",
            "osTimezone": "Asia/Tokyo",
            "appId": "tv.abema",
            "appVersion": "3.27.1",
        }
        auth_header = {"Authorization": f"Bearer {self.usertoken}"}
        res = self._session.http.get(self._MEDIATOKEN_API, params=params, headers=auth_header)
        jsonres = self._session.http.json(res, schema=self._MEDIATOKEN_SCHEMA)
        mediatoken = jsonres["token"]

        res = self._session.http.post(self._LICENSE_API, params={"t": mediatoken}, json={"kv": "a", "lt": ticket})
        jsonres = self._session.http.json(res, schema=self._LICENSE_SCHEMA)
        cid = jsonres["cid"]
        k = jsonres["k"]

        res = sum(self.STRTABLE.find(k[i]) * (58 ** (len(k) - 1 - i)) for i in range(len(k)))

        encvideokey = struct.pack(">QQ", res >> 64, res & 0xFFFFFFFFFFFFFFFF)

        # HKEY:
        # RC4KEY = unhexlify('DB98A8E7CECA3424D975280F90BD03EE')
        # RC4DATA = unhexlify(b'D4B718BBBA9CFB7D0192A58F9E2D146A'
        #                     b'FC5DB29E4352DE05FC4CF2C1005804BB')
        # rc4 = ARC4.new(RC4KEY)
        # HKEY = rc4.decrypt(RC4DATA)
        h = hmac.new(unhexlify(self.HKEY), (cid + self.deviceid).encode("utf-8"), digestmod=hashlib.sha256)
        enckey = h.digest()

        aes = AES.new(enckey, AES.MODE_ECB)
        return aes.decrypt(encvideokey)

    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None):
        resp = Response()
        resp.status_code = 200
        ticket = re.findall(r"abematv-license://(.*)", request.url)[0]
        resp._content = self._get_videokey_from_ticket(ticket)
        return resp

    def close(self):
        return


@pluginmatcher(
    name="onair",
    pattern=re.compile(r"https?://abema\.tv/now-on-air/(?P<onair>[^?]+)"),
)
@pluginmatcher(
    name="episode",
    pattern=re.compile(r"https?://abema\.tv/video/episode/(?P<episode>[^?]+)"),
)
@pluginmatcher(
    name="slots",
    pattern=re.compile(r"https?://abema\.tv/channels/.+?/slots/(?P<slots>[^?]+)"),
)
class AbemaTV(Plugin):
    _CHANNEL = "https://api.abema.io/v1/channels"

    _USER_API = "https://api.abema.io/v1/users"

    _PRGM_API = "https://api.abema.io/v1/video/programs/{0}"

    _SLOTS_API = "https://api.abema.io/v1/media/slots/{0}"

    _PRGM3U8 = "https://vod-abematv.akamaized.net/program/{0}/playlist.m3u8"

    _SLOTM3U8 = "https://vod-abematv.akamaized.net/slot/{0}/playlist.m3u8"

    SECRETKEY = (
        b"v+Gjs=25Aw5erR!J8ZuvRrCx*rGswhB&qdHd_SYerEWdU&a?3DzN9B"
        + b"Rbp5KwY4hEmcj5#fykMjJ=AuWz5GSMY-d@H7DMEh3M@9n2G552Us$$"
        + b"k9cD=3TxwWe86!x#Zyhe"
    )

    _USER_SCHEMA = validate.Schema({"profile": {"userId": str}, "token": str})

    _CHANNEL_SCHEMA = validate.Schema({
        "channels": [
            {
                "id": str,
                "name": str,
                "playback": {
                    validate.optional("dash"): str,
                    "hls": str,
                },
            },
        ],
    })

    _PRGM_SCHEMA = validate.Schema({"terms": [{validate.optional("onDemandType"): int}]})

    _SLOT_SCHEMA = validate.Schema({"slot": {"flags": {validate.optional("timeshiftFree"): bool}}})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session.http.headers.update({"User-Agent": useragents.CHROME})

    def _generate_applicationkeysecret(self, deviceid):
        deviceid = deviceid.encode("utf-8")  # for python3
        # plus 1 hour and drop minute and secs
        # for python3 : floor division
        ts_1hour = (int(time.time()) + 60 * 60) // 3600 * 3600
        time_struct = time.gmtime(ts_1hour)
        ts_1hour_str = str(ts_1hour).encode("utf-8")

        h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
        h.update(self.SECRETKEY)
        tmp = h.digest()
        for _ in range(time_struct.tm_mon):
            h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
            h.update(tmp)
            tmp = h.digest()
        h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
        h.update(urlsafe_b64encode(tmp).rstrip(b"=") + deviceid)
        tmp = h.digest()
        for _ in range(time_struct.tm_mday % 5):
            h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
            h.update(tmp)
            tmp = h.digest()

        h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
        h.update(urlsafe_b64encode(tmp).rstrip(b"=") + ts_1hour_str)
        tmp = h.digest()

        for _ in range(time_struct.tm_hour % 5):  # utc hour
            h = hmac.new(self.SECRETKEY, digestmod=hashlib.sha256)
            h.update(tmp)
            tmp = h.digest()

        return urlsafe_b64encode(tmp).rstrip(b"=").decode("utf-8")

    def _is_playable(self, vtype, vid):
        auth_header = {"Authorization": f"Bearer {self.usertoken}"}
        if vtype == "episode":
            res = self.session.http.get(self._PRGM_API.format(vid), headers=auth_header)
            jsonres = self.session.http.json(res, schema=self._PRGM_SCHEMA)
            playable = False
            for item in jsonres["terms"]:
                if item.get("onDemandType", False) == 3:
                    playable = True
            return playable
        elif vtype == "slots":
            res = self.session.http.get(self._SLOTS_API.format(vid), headers=auth_header)
            jsonres = self.session.http.json(res, schema=self._SLOT_SCHEMA)
            return jsonres["slot"]["flags"].get("timeshiftFree", False) is True

    def _get_streams(self):
        deviceid = str(uuid.uuid4())
        appkeysecret = self._generate_applicationkeysecret(deviceid)
        json_data = {"deviceId": deviceid, "applicationKeySecret": appkeysecret}
        res = self.session.http.post(self._USER_API, json=json_data)
        jsonres = self.session.http.json(res, schema=self._USER_SCHEMA)
        self.usertoken = jsonres["token"]  # for authorzation

        if self.matches["onair"]:
            onair = self.match["onair"]
            if onair == "news-global":
                self._CHANNEL = update_qsd(self._CHANNEL, {"division": "1"})
            res = self.session.http.get(self._CHANNEL)
            jsonres = self.session.http.json(res, schema=self._CHANNEL_SCHEMA)
            channels = jsonres["channels"]
            for channel in channels:
                if onair == channel["id"]:
                    break
            else:
                raise NoStreamsError
            playlisturl = channel["playback"]["hls"]
        elif self.matches["episode"]:
            episode = self.match["episode"]
            if not self._is_playable("episode", episode):
                log.error("Premium stream is not playable")
                return {}
            playlisturl = self._PRGM3U8.format(episode)
        elif self.matches["slots"]:
            slots = self.match["slots"]
            if not self._is_playable("slots", slots):
                log.error("Premium stream is not playable")
                return {}
            playlisturl = self._SLOTM3U8.format(slots)

        log.debug(f"URL={playlisturl}")

        # hook abematv private protocol
        self.session.http.mount("abematv-license://", AbemaTVLicenseAdapter(self.session, deviceid, self.usertoken))

        return AbemaTVHLSStream.parse_variant_playlist(self.session, playlisturl)


__plugin__ = AbemaTV



"""
$description TV network with multiple live TV channels broadcasting across various Latin American countries.
$url antena7.com.do
$url atv.pe
$url c9n.com.py
$url canal10.com.ni
$url canal12.com.sv
$url chapintv.com
$url elnueve.com.ar
$url redbolivision.tv.bo
$url repretel.com
$url rts.com.ec
$url snt.com.py
$url tvc.com.ec
$url vtv.com.hn
$type live
$region various
"""

import re
import time

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


@pluginmatcher(
    name="antena7",
    pattern=re.compile(r"https?://(?:www\.)?antena7\.com\.do/en-?vivo-canal-?\d{1,2}[/#]?$"),
)
@pluginmatcher(
    name="atv",
    pattern=re.compile(r"https?://(?:www\.)?atv\.pe/envivo-atv(?:mas)?/?"),
)
@pluginmatcher(
    name="c9n",
    pattern=re.compile(r"https?://(?:www\.)?c9n\.com\.py/envivo/?"),
)
@pluginmatcher(
    name="canal10",
    pattern=re.compile(r"https?://(?:www\.)?canal10\.com\.ni/envivo/?"),
)
@pluginmatcher(
    name="canal12",
    pattern=re.compile(r"https?://(?:www\.)?canal12\.com\.sv/envivo/?"),
)
@pluginmatcher(
    name="chapintv",
    pattern=re.compile(r"https?://(?:www\.)?chapintv\.com/envivo-canal-\d{1,2}/?"),
)
@pluginmatcher(
    name="elnueve",
    pattern=re.compile(r"https?://(?:www\.)?elnueve\.com\.ar/en-vivo/?"),
)
@pluginmatcher(
    name="redbolivision",
    pattern=re.compile(r"https?://(?:www\.)?redbolivision\.tv\.bo/(?:envivo-canal-?\d{1,2}|upptv)/?"),
)
@pluginmatcher(
    name="repretel",
    pattern=re.compile(r"https?://(?:www\.)?repretel\.com/en-?vivo(?:-canal-?\d{1,2})?/?"),
)
@pluginmatcher(
    name="rts",
    pattern=re.compile(r"https?://(?:www\.)?rts\.com\.ec/envivo/?"),
)
@pluginmatcher(
    name="snt",
    pattern=re.compile(r"https?://(?:www\.)?snt\.com\.py/envivo/?"),
)
@pluginmatcher(
    name="tvc",
    pattern=re.compile(r"https?://(?:www\.)?tvc\.com\.ec/envivo/?"),
)
@pluginmatcher(
    name="vtv",
    pattern=re.compile(r"https?://(?:www\.)?vtv\.com\.hn/envivo/?"),
)
class Albavision(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._page = None

    @property
    def page(self):
        if self._page is None:
            self._page = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.parse_html(),
                ),
            )
        return self._page

    def _is_token_based_site(self):
        schema = validate.Schema(
            validate.xml_xpath_string(".//script[contains(text(), 'jQuery.get')]/text()"),
        )
        is_token_based_site = schema.validate(self.page) is not None
        log.debug(f"is_token_based_site={is_token_based_site}")
        return is_token_based_site

    def _get_live_url(self):
        schema = validate.Schema(
            validate.xml_xpath_string(".//script[contains(text(), 'LIVE_URL')]/text()"),
            validate.none_or_all(
                re.compile(r"""LIVE_URL\s*=\s*(?P<q>['"])(?P<url>.+?)(?P=q)"""),
                validate.none_or_all(
                    validate.get("url"),
                    validate.url(),
                ),
            ),
        )
        live_url = schema.validate(self.page)
        log.debug(f"live_url={live_url}")
        return live_url

    def _get_token_req_url(self):
        schema = validate.Schema(
            validate.xml_xpath_string(".//script[contains(text(), 'LIVE_URL')]/text()"),
            validate.none_or_all(
                re.compile(r"""jQuery\.get\s*\((?P<q>['"])(?P<token>.+?)(?P=q)"""),
                validate.none_or_all(
                    validate.get("token"),
                    validate.url(),
                ),
            ),
        )
        token_req_host = schema.validate(self.page)
        log.debug(f"token_req_host={token_req_host}")

        schema = validate.Schema(
            validate.xml_xpath_string(".//script[contains(text(), 'LIVE_URL')]/text()"),
            validate.none_or_all(
                re.compile(r"""Math\.floor\(Date\.now\(\)\s*/\s*3600000\),\s*(?P<q>['"])(?P<token>.+?)(?P=q)"""),
                validate.none_or_all(validate.get("token")),
            ),
        )
        token_req_str = schema.validate(self.page)
        log.debug(f"token_req_str={token_req_str}")
        if not token_req_str:
            return

        date = int(time.time() // 3600)
        token_req_token = self.transform_token(token_req_str, date) or self.transform_token(token_req_str, date - 1)

        if token_req_host and token_req_token:
            return update_qsd(token_req_host, {"rsk": token_req_token})

    def _get_token(self):
        if not self._is_token_based_site():
            return

        token_req_url = self._get_token_req_url()
        if not token_req_url:
            return

        res = self.session.http.get(
            token_req_url,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "success": bool,
                    validate.optional("error"): int,
                    validate.optional("token"): str,
                },
            ),
        )

        if not res["success"]:
            if res["error"]:
                log.error(f"Token request failed with error: {res['error']}")
            else:
                log.error("Token request failed")
            return

        if not res["token"]:
            log.error("Token not found in response")
            return
        token = res["token"]
        log.debug(f"token={token}")
        return token

    @staticmethod
    def transform_token(token_in, date):
        token_out = list(token_in)
        offset = len(token_in)
        for i in range(offset - 1, -1, -1):
            p = (i * date) % offset
            # swap chars at p and i
            token_out[i], token_out[p] = token_out[p], token_out[i]
        token_out = "".join(token_out)
        if token_out.endswith("OK"):
            return token_out[:-2]
        else:
            log.error(f"Invalid site token: {token_in} => {token_out}")

    def _get_streams(self):
        live_url = self._get_live_url()
        if not live_url:
            log.info("This stream may be off-air or not available in your country")
            return

        token = self._get_token()
        if token:
            live_url = update_qsd(live_url, {"iut": token})

        return HLSStream.parse_variant_playlist(self.session, live_url)


__plugin__ = Albavision



from __future__ import annotations

import io
import json
from typing import TYPE_CHECKING

from streamlink.logger import getLogger


if TYPE_CHECKING:
    from streamlink.session import Streamlink


log = getLogger(__name__)


class Stream:
    """
    This is a base class that should be inherited when implementing
    different stream types. Should only be created by plugins.
    """

    __shortname__ = "stream"

    @classmethod
    def shortname(cls):
        return cls.__shortname__

    def __init__(self, session: Streamlink):
        """
        :param session: Streamlink session instance
        """

        self.session: Streamlink = session

    def __repr__(self):
        params = [repr(self.shortname())]
        for method in self.to_url, self.to_manifest_url:
            try:
                params.append(repr(method()))
            except TypeError:
                pass

        return f"<{self.__class__.__name__} [{', '.join(params)}]>"

    def __json__(self):  # noqa: PLW3201
        return dict(type=self.shortname())

    @property
    def json(self):
        obj = self.__json__()
        return json.dumps(obj)

    def to_url(self):
        raise TypeError(f"<{self.__class__.__name__} [{self.shortname()}]> cannot be translated to a URL")

    def to_manifest_url(self):
        raise TypeError(f"<{self.__class__.__name__} [{self.shortname()}]> cannot be translated to a manifest URL")

    def open(self) -> StreamIO:
        """
        Attempts to open a connection to the stream.
        Returns a file-like object that can be used to read the stream data.

        :raises StreamError: on failure
        """

        raise NotImplementedError


class StreamIO(io.IOBase):
    pass


__all__ = ["Stream", "StreamIO"]



"""
$description Live TV channels from RTVS, a Slovak public, state-owned broadcaster.
$url rtvs.sk
$type live
$region Slovakia
"""

import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    re.compile(r"https?://www\.rtvs\.sk/televizia/(?:live-|sport)"),
)
class Rtvs(Plugin):
    def _get_streams(self):
        channel = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//iframe[@id='player_live']//@src"),
                validate.url(path=validate.startswith("/embed/live/")),
                validate.transform(lambda embed: urlparse(embed).path[len("/embed/live/") :]),
            ),
        )
        if not channel:
            return

        videos = self.session.http.get(
            "https://www.rtvs.sk/json/live5f.json",
            params={
                "c": channel,
                "b": "mozilla",
                "p": "win",
                "f": "0",
                "d": "1",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "clip": {
                        "sources": [
                            {
                                "src": validate.url(),
                                "type": str,
                            },
                        ],
                    },
                },
                validate.get(("clip", "sources")),
                validate.filter(lambda n: n["type"] == "application/x-mpegurl"),
            ),
        )
        for video in videos:
            return HLSStream.parse_variant_playlist(self.session, video["src"])


__plugin__ = Rtvs



"""
$description Global live-streaming and video hosting social platform owned by Amazon.
$url twitch.tv
$type live, vod
$webbrowser Required for getting a new :ref:`client-integrity token <cli/plugins/twitch:Client-integrity token>`.
$metadata id
$metadata author
$metadata category
$metadata title
$notes See the :ref:`Authentication <cli/plugins/twitch:Authentication>` docs on how to prevent ads.
$notes Read more about :ref:`embedded ads <cli/plugins/twitch:Embedded ads>` here.
$notes :ref:`Higher quality streams <cli/plugins/twitch:Higher quality streams>` are supported.
$notes :ref:`Low latency streaming <cli/plugins/twitch:Low latency streaming>` is supported.
$notes Acquires a :ref:`client-integrity token <cli/plugins/twitch:Client-integrity token>` on streaming access token failure.
"""

from __future__ import annotations

import argparse
import math
import re
import sys
from collections import deque
from contextlib import suppress
from dataclasses import dataclass, replace as dataclass_replace
from datetime import timedelta
from json import dumps as json_dumps
from random import random
from typing import TYPE_CHECKING, ClassVar
from urllib.parse import urlparse

from requests.exceptions import HTTPError

from streamlink.exceptions import NoStreamsError, PluginError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.session import http_useragents
from streamlink.stream.hls import (
    M3U8,
    HLSPlaylist,
    HLSSegment,
    HLSStream,
    HLSStreamReader,
    HLSStreamWorker,
    HLSStreamWriter,
    M3U8Parser,
    Media,
    parse_tag,
)
from streamlink.stream.http import HTTPStream
from streamlink.utils.parse import parse_json, parse_qsd
from streamlink.utils.random import CHOICES_ALPHA_NUM, random_token
from streamlink.utils.times import fromtimestamp, hours_minutes_seconds_float
from streamlink.utils.url import update_qsd


if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime

    from streamlink.session import Streamlink
    from streamlink.stream.hls import DateRange


log = getLogger(__name__)

LOW_LATENCY_MAX_LIVE_EDGE = 2


@dataclass(kw_only=True)
class TwitchHLSSegment(HLSSegment):
    ad: bool
    prefetch: bool


class TwitchM3U8(M3U8[TwitchHLSSegment, HLSPlaylist]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.dateranges_ads: list[DateRange] = []


class TwitchM3U8Parser(M3U8Parser[TwitchM3U8, TwitchHLSSegment, HLSPlaylist]):
    __m3u8__: ClassVar[type[TwitchM3U8]] = TwitchM3U8
    __segment__: ClassVar[type[TwitchHLSSegment]] = TwitchHLSSegment

    @parse_tag("EXT-X-TWITCH-LIVE-SEQUENCE")
    def parse_ext_x_twitch_live_sequence(self, *_):
        # Unset discontinuity state if the previous segment was not an ad,
        # as the following segment won't be an ad
        if self.m3u8.segments and not self.m3u8.segments[-1].ad:
            self._discontinuity = False

    @parse_tag("EXT-X-TWITCH-PREFETCH")
    def parse_tag_ext_x_twitch_prefetch(self, value):
        segments = self.m3u8.segments
        if not segments:  # pragma: no cover
            return
        last = segments[-1]

        # Use the average duration of all regular segments for the duration of prefetch segments.
        # This is better than using the duration of the last segment when regular segment durations vary a lot.
        # In low latency mode, the playlist reload time is the duration of the last segment.
        duration = last.duration if last.prefetch else sum(segment.duration for segment in segments) / float(len(segments))

        # Use the last duration for extrapolating the start time of the prefetch segment, which is needed for checking
        # whether it is an ad segment and matches the parsed date ranges or not
        if not last.date:
            return
        date = last.date + timedelta(seconds=last.duration)

        # Always treat prefetch segments after a discontinuity as ad segments
        # (discontinuity tag inserted after last regular segment)
        # Don't reset discontinuity state: the date extrapolation might be inaccurate,
        # so all following prefetch segments should be considered an ad after a discontinuity
        ad = self._discontinuity or self._is_segment_ad(date)

        # Since we don't reset the discontinuity state in prefetch segments for the purpose of ad detection,
        # set the prefetch segment's discontinuity attribute based on ad transitions
        discontinuity = ad != last.ad

        segment = dataclass_replace(
            last,
            uri=self.uri(value),
            duration=duration,
            title=None,
            discontinuity=discontinuity,
            date=date,
            ad=ad,
            prefetch=True,
        )
        segments.append(segment)

    @parse_tag("EXT-X-DATERANGE")
    def parse_tag_ext_x_daterange(self, value):
        super().parse_tag_ext_x_daterange(value)
        daterange = self.m3u8.dateranges[-1]
        if self._is_daterange_ad(daterange):
            self.m3u8.dateranges_ads.append(daterange)
            log.trace("Advertisement: %r", daterange)

    def get_segment(self, uri: str, **data) -> TwitchHLSSegment:
        ad = self._is_segment_ad(self._date, self._extinf.title if self._extinf else None)
        segment: TwitchHLSSegment = super().get_segment(uri, ad=ad, prefetch=False)  # type: ignore[assignment, ty:invalid-assignment]

        # Special case where Twitch incorrectly inserts discontinuity tags between segments of the live content
        if (
            segment.discontinuity
            and not segment.ad
            and self.m3u8.segments
            and not self.m3u8.segments[-1].ad
        ):  # fmt: skip
            segment.discontinuity = False

        return segment

    def get_playlist(self, *args, **kwargs):
        streaminf = self._streaminf or {}
        playlist = super().get_playlist(*args, **kwargs)
        # backwards compatibility for stream names on Usher v2
        if not playlist.media and (name := streaminf.get("IVS-NAME")):
            is_audio_only = name in ("audio_only", "audio")  # live + VOD
            media = Media(
                uri=None,
                type="VIDEO",
                group_id=name,
                language=None,
                name=name,
                default=not is_audio_only,
                autoselect=not is_audio_only,
                forced=False,
                characteristics=None,
            )
            playlist.stream_info.video = name
            playlist.media.append(media)

        return playlist

    def _is_segment_ad(self, date: datetime | None, title: str | None = None) -> bool:
        return (
            title is not None and "Amazon" in title
            or any(self.m3u8.is_date_in_daterange(date, daterange) for daterange in self.m3u8.dateranges_ads)
        )  # fmt: skip

    @staticmethod
    def _is_daterange_ad(daterange: DateRange) -> bool:
        return (
            daterange.classname == "twitch-stitched-ad"
            or str(daterange.id or "").startswith("stitched-ad-")
        )  # fmt: skip


class TwitchHLSStreamWorker(HLSStreamWorker):
    reader: TwitchHLSStreamReader
    writer: TwitchHLSStreamWriter
    stream: TwitchHLSStream

    def __init__(self, reader, *args, **kwargs) -> None:
        self.had_content: bool = False
        self.logged_ads: deque[str] = deque(maxlen=10)
        super().__init__(reader, *args, **kwargs)
        if self.stream.low_latency:
            self.reload_time = "segment"

    def process_segments(self, playlist: TwitchM3U8):  # type: ignore[override, ty:invalid-method-override]
        # ignore prefetch segments if not LL streaming
        if not self.stream.low_latency:
            playlist.segments = [segment for segment in playlist.segments if not segment.prefetch]

        # set ad segment duration to zero, so it doesn't affect the worker's `duration` attribute
        # do it here instead of the parser because prefetch segment durations are averaged over all regular segments
        for segment in playlist.segments:
            if segment.ad:
                segment.duration = 0.0

        # check for sequences with real content
        if not self.had_content:
            self.had_content = next((True for segment in playlist.segments if not segment.ad), False)

            # When filtering ads, to check whether it's a LL stream, we need to wait for the real content to show up,
            # since playlists with only ad segments don't contain prefetch segments
            if (
                self.stream.low_latency
                and self.had_content
                and not next((True for segment in playlist.segments if segment.prefetch), False)
            ):
                log.info("This is not a low latency stream")

        # show pre-roll ads message only on the first playlist containing ads
        if self.sequence == -1 and not self.had_content:
            log.info("Waiting for pre-roll ads to finish, be patient")

        # log the duration of whole advertisement breaks
        for daterange_ads in playlist.dateranges_ads:
            if not daterange_ads.duration:  # pragma: no cover
                continue

            ads_id: str | None = (
                daterange_ads.x.get("X-TV-TWITCH-AD-COMMERCIAL-ID")
                or daterange_ads.x.get("X-TV-TWITCH-AD-ROLL-TYPE")
            )  # fmt: skip
            if not ads_id or ads_id in self.logged_ads:
                continue
            self.logged_ads.append(ads_id)

            # use Twitch's own ads duration metadata if available
            try:
                duration = math.ceil(float(daterange_ads.x.get("X-TV-TWITCH-AD-POD-FILLED-DURATION", "")))
            except ValueError:
                duration = math.ceil(daterange_ads.duration.total_seconds())

            log.info(f"Detected advertisement break of {duration} second{'s' if duration != 1 else ''}")

        return super().process_segments(playlist)


class TwitchHLSStreamWriter(HLSStreamWriter):
    reader: TwitchHLSStreamReader
    stream: TwitchHLSStream

    def should_filter_segment(self, segment: TwitchHLSSegment) -> bool:  # type: ignore[override, ty:invalid-method-override]
        return segment.ad


class TwitchHLSStreamReader(HLSStreamReader):
    __worker__ = TwitchHLSStreamWorker
    __writer__ = TwitchHLSStreamWriter

    worker: TwitchHLSStreamWorker
    writer: TwitchHLSStreamWriter
    stream: TwitchHLSStream

    def __init__(self, stream: TwitchHLSStream, **kwargs):
        log.info("Will skip ad segments")
        if stream.low_latency:
            live_edge = max(1, min(LOW_LATENCY_MAX_LIVE_EDGE, stream.session.options.get("hls-live-edge")))
            stream.session.options.set("hls-live-edge", live_edge)
            stream.session.options.set("hls-segment-stream-data", True)
            log.info(f"Low latency streaming (HLS live edge: {live_edge})")

        super().__init__(stream, **kwargs)


class TwitchHLSStream(HLSStream):
    __reader__ = TwitchHLSStreamReader
    __parser__ = TwitchM3U8Parser

    def __init__(self, *args, low_latency: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.low_latency = low_latency


class UsherService:
    SUPPORTED_CODECS_DEFAULT = ["h264"]

    def __init__(self, session: Streamlink, supported_codecs: list[str] | None = None):
        self.session = session
        self.supported_codecs = supported_codecs or self.SUPPORTED_CODECS_DEFAULT

    def _create_url(self, endpoint, **extra_params):
        url = f"https://usher.ttvnw.net{endpoint}"
        params = {
            "platform": "web",
            "p": int(random() * 999999),
            "allow_source": "true",
            "allow_audio_only": "true",
            "playlist_include_framerate": "true",
            "supported_codecs": ",".join(self.supported_codecs),
        }
        params.update(extra_params)

        req = self.session.http.prepare_new_request(url=url, params=params)

        return req.url

    def channel(self, channel: str, **extra_params) -> str:
        with suppress(PluginError):
            extra_params_debug = validate.Schema(
                validate.get("token"),
                validate.parse_json(),
                {
                    "adblock": bool,
                    "geoblock_reason": str,
                    "hide_ads": bool,
                    "server_ads": bool,
                    "show_ads": bool,
                },
            ).validate(extra_params)
            log.debug("%r", extra_params_debug)

        return self._create_url(f"/api/v2/channel/hls/{channel.lower()}.m3u8", **extra_params)

    def video(self, video_id: str, **extra_params) -> str:
        return self._create_url(f"/vod/v2/{video_id}.m3u8", **extra_params)


class TwitchAPI:
    CLIENT_ID = "kimne78kx3ncx6brgo4mv6wki5h1ko"

    def __init__(self, session, api_header=None, access_token_param=None):
        self.session = session
        self.headers = {
            "Client-ID": self.CLIENT_ID,
        }
        self.headers.update(**dict(api_header or []))
        self.access_token_params = dict(access_token_param or [])
        self.access_token_params.setdefault("playerType", "embed")
        self.access_token_params.setdefault("platform", "site")

    def call(self, data, /, *, headers=None, schema, **kwargs):
        return self.session.http.post(
            "https://gql.twitch.tv/gql",
            json=data,
            headers={
                **self.headers,
                **(headers or {}),
            },
            schema=validate.Schema(
                validate.parse_json(),
                schema,
            ),
            **kwargs,
        )

    @staticmethod
    def _gql_persisted_query(operationname, sha256hash, **variables):
        return {
            "operationName": operationname,
            "extensions": {
                "persistedQuery": {
                    "version": 1,
                    "sha256Hash": sha256hash,
                },
            },
            "variables": dict(**variables),
        }

    @staticmethod
    def parse_token(tokenstr):
        return parse_json(
            tokenstr,
            schema=validate.Schema(
                {
                    "chansub": {
                        "restricted_bitrates": validate.all(
                            [str],
                            validate.filter(lambda n: not re.match(r"(.+_)?archives|live|chunked", n)),
                        ),
                    },
                },
                validate.get(("chansub", "restricted_bitrates")),
            ),
        )

    # GraphQL API calls

    def metadata_video(self, video_id):
        query = self._gql_persisted_query(
            "VideoMetadata",
            "45111672eea2e507f8ba44d101a61862f9c56b11dee09a15634cb75cb9b9084d",
            channelLogin="",  # parameter can be empty
            videoID=video_id,
        )

        return self.call(
            query,
            schema=validate.all(
                {
                    "data": {
                        "video": {
                            "id": str,
                            "owner": {
                                "displayName": str,
                            },
                            "title": str,
                            "game": {
                                "displayName": str,
                            },
                        },
                    },
                },
                validate.get(("data", "video")),
                validate.union_get(
                    "id",
                    ("owner", "displayName"),
                    ("game", "displayName"),
                    "title",
                ),
            ),
        )

    def metadata_channel(self, channel):
        queries = [
            self._gql_persisted_query(
                "ChannelShell",
                "fea4573a7bf2644f5b3f2cbbdcbee0d17312e48d2e55f080589d053aad353f11",
                login=channel,
            ),
            self._gql_persisted_query(
                "StreamMetadata",
                "b57f9b910f8cd1a4659d894fe7550ccc81ec9052c01e438b290fd66a040b9b93",
                channelLogin=channel,
                includeIsDJ=True,
            ),
        ]

        return self.call(
            queries,
            schema=validate.all(
                validate.list(
                    validate.all(
                        {
                            "data": {
                                "userOrError": {
                                    "displayName": str,
                                },
                            },
                        },
                    ),
                    validate.all(
                        {
                            "data": {
                                "user": {
                                    "lastBroadcast": {
                                        "title": str,
                                    },
                                    "stream": {
                                        "id": str,
                                        "game": {
                                            "name": str,
                                        },
                                    },
                                },
                            },
                        },
                    ),
                ),
                validate.union_get(
                    (1, "data", "user", "stream", "id"),
                    (0, "data", "userOrError", "displayName"),
                    (1, "data", "user", "stream", "game", "name"),
                    (1, "data", "user", "lastBroadcast", "title"),
                ),
            ),
        )

    def metadata_clips(self, clipname):
        query = self._gql_persisted_query(
            "ShareClipRenderStatus",
            "1844261bb449fa51e6167040311da4a7a5f1c34fe71c71a3e0c4f551bc30c698",
            slug=clipname,
        )

        return self.call(
            query,
            schema=validate.all(
                {
                    "data": {
                        "clip": {
                            "id": str,
                            "broadcaster": {"displayName": str},
                            "game": {"name": str},
                            "title": str,
                        },
                    },
                },
                validate.get(("data", "clip")),
                validate.union_get(
                    "id",
                    ("broadcaster", "displayName"),
                    ("game", "name"),
                    "title",
                ),
            ),
        )

    def access_token(self, is_live, channel_or_vod, client_integrity: tuple[str, str] | None = None):
        query = self._gql_persisted_query(
            "PlaybackAccessToken",
            "ed230aa1e33e07eebb8928504583da78a5173989fadfb1ac94be06a04f3cdbe9",
            isLive=is_live,
            login=channel_or_vod if is_live else "",
            isVod=not is_live,
            vodID=channel_or_vod if not is_live else "",
            **self.access_token_params,
        )
        subschema = validate.none_or_all(
            {
                "value": str,
                "signature": str,
            },
            validate.union_get("signature", "value"),
        )

        headers = {
            # https://github.com/streamlink/streamlink/issues/6574
            "User-Agent": http_useragents.DEFAULT,
        }
        if client_integrity:
            headers["Device-Id"], headers["Client-Integrity"] = client_integrity

        return self.call(
            query,
            acceptable_status=(200, 400, 401, 403),
            headers=headers,
            schema=validate.any(
                validate.all(
                    {"errors": [{"message": str}]},
                    validate.get(("errors", 0, "message")),
                    validate.transform(lambda data: ("error", None, data)),
                ),
                validate.all(
                    {"error": str, "message": str},
                    validate.union_get("error", "message"),
                    validate.transform(lambda data: ("error", *data)),
                ),
                validate.all(
                    {
                        "data": validate.any(
                            validate.all(
                                {"streamPlaybackAccessToken": subschema},
                                validate.get("streamPlaybackAccessToken"),
                            ),
                            validate.all(
                                {"videoPlaybackAccessToken": subschema},
                                validate.get("videoPlaybackAccessToken"),
                            ),
                        ),
                    },
                    validate.get("data"),
                    validate.transform(lambda data: ("token", *data) if data is not None else ("token", None, None)),
                ),
            ),
        )

    def clips(self, clipname):
        query = self._gql_persisted_query(
            "VideoAccessToken_Clip",
            "993d9a5131f15a37bd16f32342c44ed1e0b1a9b968c6afdb662d2cddd595f6c5",
            slug=clipname,
            platform="web",
        )

        return self.call(
            query,
            schema=validate.all(
                {
                    "data": {
                        "clip": validate.none_or_all(
                            {
                                "playbackAccessToken": {
                                    "signature": str,
                                    "value": str,
                                },
                                "videoQualities": validate.all(
                                    [
                                        {
                                            "frameRate": validate.transform(int),
                                            "quality": str,
                                            "sourceURL": validate.any("", validate.url()),
                                        },
                                    ],
                                    validate.filter(lambda clip: clip["sourceURL"]),
                                    validate.map(
                                        lambda clip: (
                                            f"{clip['quality']}p{clip['frameRate']}",
                                            clip["sourceURL"],
                                        ),
                                    ),
                                ),
                            },
                            validate.union_get(
                                ("playbackAccessToken", "signature"),
                                ("playbackAccessToken", "value"),
                                "videoQualities",
                            ),
                        ),
                    },
                },
                validate.get(("data", "clip")),
            ),
        )


class TwitchClientIntegrity:
    URL_P_SCRIPT = "https://k.twitchcdn.net/149e9513-01fa-4fb0-aad4-566afd725d1b/2d206a39-8ed7-437e-a3be-862e0f06eea3/p.js"

    # language=javascript
    JS_INTEGRITY_TOKEN = """
    // noinspection JSIgnoredPromiseFromCall
    new Promise((resolve, reject) => {
        function configureKPSDK() {
            // noinspection JSUnresolvedVariable,JSUnresolvedFunction
            window.KPSDK.configure([{
                "protocol": "https:",
                "method": "POST",
                "domain": "gql.twitch.tv",
                "path": "/integrity"
            }]);
        }

        async function fetchIntegrity() {
            // noinspection JSUnresolvedReference
            const headers = Object.assign(HEADERS, {"x-device-id": "DEVICE_ID"});
            // window.fetch gets overridden and the patched function needs to be used
            const resp = await window.fetch("https://gql.twitch.tv/integrity", {
                "headers": headers,
                "body": null,
                "method": "POST",
                "mode": "cors",
                "credentials": "omit"
            });

            if (resp.status !== 200) {
                throw new Error(`Unexpected integrity response status code ${resp.status}`);
            }

            return JSON.stringify(await resp.json());
        }

        document.addEventListener("kpsdk-load", configureKPSDK, {once: true});
        document.addEventListener("kpsdk-ready", () => fetchIntegrity().then(resolve, reject), {once: true});

        const script = document.createElement("script");
        script.addEventListener("error", reject);
        script.src = "SCRIPT_SOURCE";
        document.body.appendChild(script);
    });
    """

    @classmethod
    def acquire(
        cls,
        session: Streamlink,
        channel: str,
        headers: Mapping[str, str],
        device_id: str,
    ) -> tuple[str, int] | None:
        from streamlink.compat import BaseExceptionGroup  # noqa: PLC0415
        from streamlink.webbrowser.cdp import CDPClient, CDPClientSession, devtools  # noqa: PLC0415, TC001

        url = f"https://www.twitch.tv/{channel}"
        js_get_integrity_token = cls.JS_INTEGRITY_TOKEN \
            .replace("SCRIPT_SOURCE", cls.URL_P_SCRIPT) \
            .replace("HEADERS", json_dumps(headers)) \
            .replace("DEVICE_ID", device_id)  # fmt: skip
        eval_timeout = session.get_option("webbrowser-timeout")
        # noinspection PyUnusedLocal
        client_integrity: str | None = None

        async def on_main(client_session: CDPClientSession, request: devtools.fetch.RequestPaused):
            async with client_session.alter_request(request) as cm:
                cm.body = "<!doctype html>"

        async def acquire_client_integrity_token(client: CDPClient):
            client_session: CDPClientSession
            async with client.session() as client_session:
                client_session.add_request_handler(on_main, url_pattern=url, on_request=True)
                async with client_session.navigate(url) as frame_id:
                    await client_session.loaded(frame_id)
                    return await client_session.evaluate(js_get_integrity_token, timeout=eval_timeout)

        try:
            client_integrity = CDPClient.launch(session, acquire_client_integrity_token)
        except BaseExceptionGroup:
            log.exception("Failed acquiring client integrity token")
        except Exception as err:
            log.error(err)

        if not client_integrity:
            return None

        schema = validate.Schema(
            validate.parse_json(),
            {"token": str, "expiration": int},
            validate.union_get("token", "expiration"),
        )
        token, expiration = schema.validate(client_integrity)

        return token, expiration / 1000


@pluginmatcher(
    name="player",
    pattern=re.compile(
        r"https?://player\.twitch\.tv/\?.+",
    ),
)
@pluginmatcher(
    name="clip",
    pattern=re.compile(
        r"https?://(?:clips\.twitch\.tv|(?:[\w-]+\.)?twitch\.tv/(?:[\w-]+/)?clip)/(?P<clip_id>[^/?]+)",
    ),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(
        r"https?://(?:[\w-]+\.)?twitch\.tv/(?:[\w-]+/)?v(?:ideos?)?/(?P<video_id>\d+)",
    ),
)
@pluginmatcher(
    name="live",
    pattern=re.compile(
        r"https?://(?:(?!clips\.)[\w-]+\.)?twitch\.tv/(?P<channel>(?!v(?:ideos?)?/|clip/)[^/?]+)/?(?:\?|$)",
    ),
)
@pluginargument(
    "disable-ads",
    action="store_true",
    help=argparse.SUPPRESS,
)
@pluginargument(
    "disable-hosting",
    action="store_true",
    help=argparse.SUPPRESS,
)
@pluginargument(
    "disable-reruns",
    action="store_true",
    help=argparse.SUPPRESS,
)
@pluginargument(
    "low-latency",
    action="store_true",
    help="""
        Enables low latency streaming by prefetching HLS segments.
        Sets --hls-segment-stream-data to true and --hls-live-edge to 2, if it is higher.
        Reducing --hls-live-edge to `1` will result in the lowest latency possible, but will most likely cause buffering.

        In order to achieve true low latency streaming during playback, the player's caching/buffering settings will
        need to be adjusted and reduced to a value as low as possible, but still high enough to not cause any buffering.
        This depends on the stream's bitrate and the quality of the connection to Twitch's servers. Please refer to the
        player's own documentation for the required configuration. Player parameters can be set via --player-args.

        Note: Low latency streams have to be enabled by the broadcasters on Twitch themselves.
        Regular streams can cause buffering issues with this option enabled due to the reduced --hls-live-edge value.
    """,
)
@pluginargument(
    "supported-codecs",
    metavar="CODECS",
    type="comma_list_filter",
    type_kwargs={
        "acceptable": ["h264", "h265", "av1"],
        "unique": True,
    },
    default=["h264"],
    help="""
        A comma-separated list of codec names which signals Twitch the client's stream codec preference.
        Which streams and which codecs are available depends on the specific channel and broadcast.

        Default is "h264".

        Supported codecs are h264, h265 and av1. Set to "h264,h265,av1" to enable all codecs.

        Higher quality streams may only be available by enabling h265 or av1.
        Lower quality streams which are re-encoded on Twitch's end may still be h264, even if not requested.
    """,
)
@pluginargument(
    "api-header",
    metavar="KEY=VALUE",
    type="keyvalue",
    action="append",
    help="""
        A header to add to each Twitch API HTTP request.

        Can be repeated to add multiple headers.

        Useful for adding authentication data that can prevent ads. See the plugin-specific documentation for more information.
    """,
)
@pluginargument(
    "access-token-param",
    metavar="KEY=VALUE",
    type="keyvalue",
    action="append",
    help="""
        A parameter to add to the API request for acquiring the streaming access token.

        Can be repeated to add multiple parameters.
    """,
)
@pluginargument(
    "force-client-integrity",
    action="store_true",
    help="Don't attempt requesting the streaming access token without a client-integrity token.",
)
@pluginargument(
    "purge-client-integrity",
    action="store_true",
    help="Purge cached Twitch client-integrity token and acquire a new one.",
)
class Twitch(Plugin):
    _CACHE_KEY_CLIENT_INTEGRITY = "client-integrity"

    api: TwitchAPI
    usher: UsherService

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        if stream == "source":
            return sys.maxsize, stream
        return super().stream_weight(stream)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        params = parse_qsd(urlparse(self.url).query)

        self.channel = self.match["channel"] if self.matches["live"] else None
        self.video_id = self.match["video_id"] if self.matches["vod"] else None
        self.clip_id = self.match["clip_id"] if self.matches["clip"] else None

        if self.matches["player"]:
            self.channel = params.get("channel")
            self.video_id = params.get("video")

        try:
            self.time_offset = hours_minutes_seconds_float(params.get("t", "0"))
        except ValueError:
            self.time_offset = 0

        self.api = TwitchAPI(
            session=self.session,
            api_header=self.get_option("api-header"),
            access_token_param=self.get_option("access-token-param"),
        )
        self.usher = UsherService(
            session=self.session,
            supported_codecs=self.get_option("supported-codecs"),
        )

        self._checked_metadata = False

        def method_factory(parent_method):
            def inner():
                if not self._checked_metadata:
                    self._checked_metadata = True
                    self._get_metadata()
                return parent_method()

            return inner

        parent = super()
        for metadata in "id", "author", "category", "title":
            method = f"get_{metadata}"
            setattr(self, method, method_factory(getattr(parent, method)))

    def _get_metadata(self):
        try:
            if self.video_id:
                data = self.api.metadata_video(self.video_id)
            elif self.clip_id:
                data = self.api.metadata_clips(self.clip_id)
            elif self.channel:
                data = self.api.metadata_channel(self.channel)
            else:  # pragma: no cover
                return
            self.id, self.author, self.category, self.title = data
        except (PluginError, TypeError):
            pass

    def _client_integrity_token(self, channel: str) -> tuple[str, str] | None:
        if self.options.get("purge-client-integrity"):
            log.info("Removing cached client-integrity token...")
            self.cache.set(self._CACHE_KEY_CLIENT_INTEGRITY, None, 0)

        client_integrity = self.cache.get(self._CACHE_KEY_CLIENT_INTEGRITY)
        if client_integrity and isinstance(client_integrity, list) and len(client_integrity) == 2:
            log.info("Using cached client-integrity token")
            device_id, token = client_integrity
        else:
            log.info("Acquiring new client-integrity token...")
            device_id = random_token(32, CHOICES_ALPHA_NUM)
            client_integrity = TwitchClientIntegrity.acquire(
                self.session,
                channel,
                self.api.headers,
                device_id,
            )
            if not client_integrity:
                log.warning("No client-integrity token acquired")
                return None

            token, expiration = client_integrity
            self.cache.set(self._CACHE_KEY_CLIENT_INTEGRITY, [device_id, token], expires_at=fromtimestamp(expiration))

        return device_id, token

    def _access_token(self, is_live, channel_or_vod):
        response = ""
        data = (None, None)

        # if live, try without a client-integrity token first (the web player did the same on 2023-05-31)
        # if not live, we don't need a client-integrity token
        if not is_live or not self.options.get("force-client-integrity"):
            response, *data = self.api.access_token(is_live, channel_or_vod)

        # if live and the previous API response was erroneous, try again with a client-integrity token
        if is_live and response != "token":
            client_integrity = self._client_integrity_token(channel_or_vod)
            response, *data = self.api.access_token(is_live, channel_or_vod, client_integrity)

        # unknown API response error: abort
        if response != "token":
            error, message = data
            raise PluginError(f"{error or 'Error'}: {message or 'Unknown error'}")

        # access token response was empty: stream is offline or channel doesn't exist
        elif data[0] is None:
            raise NoStreamsError

        sig, token = data
        try:
            restricted_bitrates = self.api.parse_token(token)
        except PluginError:
            restricted_bitrates = []

        return sig, token, restricted_bitrates

    def _get_hls_streams_live(self, channel: str):
        # only get the token once the channel has been resolved
        log.debug(f"Getting live HLS streams for {channel}")
        self.session.http.headers.update({
            "referer": "https://player.twitch.tv",
            "origin": "https://player.twitch.tv",
        })
        sig, token, restricted_bitrates = self._access_token(True, channel)
        url = self.usher.channel(channel, sig=sig, token=token, fast_bread=True)

        return self._get_hls_streams(url, restricted_bitrates)

    def _get_hls_streams_video(self, video_id: str):
        log.debug(f"Getting HLS streams for video ID {self.video_id}")
        sig, token, restricted_bitrates = self._access_token(False, video_id)
        url = self.usher.video(video_id, nauthsig=sig, nauth=token)

        # If the stream is a VOD that is still being recorded, the stream should start at the beginning of the recording
        return self._get_hls_streams(url, restricted_bitrates, force_restart=True)

    def _get_hls_streams(self, url: str, restricted_bitrates: list[str], **extra_params):
        try:
            streams = TwitchHLSStream.parse_variant_playlist(
                self.session,
                url,
                start_offset=self.time_offset,
                # Check if the media playlists are accessible:
                # This is a workaround for checking the GQL API for the channel's live status,
                # which can be delayed by up to a minute.
                check_streams=True,
                low_latency=self.get_option("low-latency"),
                **extra_params,
            )
        except OSError as err:
            # TODO: fix the "err" attribute set by HTTPSession.request()
            orig = getattr(err, "err", None)
            if isinstance(orig, HTTPError) and orig.response.status_code >= 400:
                # The playlist's error response may include JSON data with an error message
                with suppress(PluginError):
                    error = validate.Schema(
                        validate.parse_json(),
                        [
                            {
                                "type": "error",
                                "error": str,
                            },
                        ],
                        validate.get((0, "error")),
                    ).validate(orig.response.text)
                    # Only log error messages if the channel is actually live
                    if self.get_id():
                        log.error(error or "Could not access HLS playlist")
                # Don't raise and simply return no streams on 4xx/5xx playlist responses
                return None
            raise PluginError(err) from err

        for name in restricted_bitrates:
            if name not in streams:
                log.warning(f"The quality '{name}' is not available since it requires a subscription.")

        return streams

    def _get_clips(self, clip_id: str):
        data = self.api.clips(clip_id)
        if not data:
            return
        sig, token, streams = data
        for quality, stream in streams:
            yield quality, HTTPStream(self.session, update_qsd(stream, {"sig": sig, "token": token}))

    def _get_streams(self):
        if self.video_id:
            return self._get_hls_streams_video(self.video_id)
        elif self.clip_id:
            return self._get_clips(self.clip_id)
        elif self.channel:
            return self._get_hls_streams_live(self.channel)
        return None


__plugin__ = Twitch



"""
$description Live TV channels and VODs from TVP, a Polish public, state-owned broadcaster.
$url vod.tvp.pl
$url tvp.info
$url sport.tvp.pl
$type live, vod
$metadata id
$metadata title
$region Poland
$notes Some live streams and VODs may be geo-restricted. Authentication is not supported.
"""

from __future__ import annotations

import re

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(
        r"https?://vod\.tvp\.pl/live,\d+/[^,]+,(?P<channel_id>\d+)(?:/.+,(?P<show_id>\d+)$)?",
    ),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(
        r"https?://vod\.tvp\.pl/(?!live,)[^/]+/.+,(?P<vod_id>\d+)$",
    ),
)
@pluginmatcher(
    name="tvp_info",
    pattern=re.compile(
        r"https?://(?:www\.)?tvp\.info/",
    ),
)
@pluginmatcher(
    name="tvp_sport",
    pattern=re.compile(
        r"https?://sport\.tvp\.pl/(?P<stream_id>\d+)/.+",
    ),
)
class TVP(Plugin):
    _URL_VOD = "https://vod.tvp.pl/api/products/{vod_id}/videos/playlist"
    _URL_INFO_API_TOKEN = "https://api.tvp.pl/tokenizer/token/{token}"
    _URL_INFO_API_NEWS = "https://www.tvp.info/api/info/news?device=www&id={id}"
    _URL_INFO_API_BLOCK = "https://www.tvp.info/api/info/block?device=www&id={id}"
    _URL_INFO_API_BLOCK_LIST = "https://www.tvp.info/api/info/block/list?device=www&id={id}"

    def _get_formats_from_api(self, token):
        is_geo_blocked, self.title, formats = self.session.http.get(
            self._URL_INFO_API_TOKEN.format(token=token),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "status": "OK",
                    "isGeoBlocked": bool,
                    "title": str,
                    "formats": [
                        validate.all(
                            {
                                "mimeType": str,
                                "url": validate.url(),
                            },
                            validate.union_get("mimeType", "url"),
                        ),
                    ],
                },
                validate.union_get("isGeoBlocked", "title", "formats"),
            ),
        )
        if is_geo_blocked:
            log.error("The content is not available in your region")
            raise NoStreamsError

        for mime_type, url in formats:
            if mime_type == "application/x-mpegurl":
                yield from HLSStream.parse_variant_playlist(self.session, url).items()
            if mime_type == "application/dash+xml":
                yield from DASHStream.parse_manifest(self.session, url).items()

    def _get_vod(self, vod_id: str, video_type: str = "MOVIE"):
        data = self.session.http.get(
            self._URL_VOD.format(vod_id=vod_id),
            params={
                "platform": "BROWSER",
                "videoType": video_type,
            },
            acceptable_status=(200, 403),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {"code": str},
                    validate.all(
                        {
                            "sources": {
                                validate.optional("HLS"): [
                                    {
                                        "src": validate.url(),
                                    },
                                ],
                            },
                        },
                        validate.get("sources"),
                    ),
                ),
            ),
        )

        if data.get("code") == "GEOIP_FILTER_FAILED":
            log.error("The content is not available in your region")
            return

        if data.get("code") == "ITEM_NOT_PAID":
            log.error("The content is not accessible")
            return

        if data.get("HLS"):
            return HLSStream.parse_variant_playlist(self.session, data["HLS"][0]["src"])

    def _get_tvp_info(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'window.__pageSettings')][1]/text()"),
                validate.none_or_all(
                    str,
                    validate.regex(re.compile(r"window\.__pageSettings\s*=\s*(?P<json>{.*?})\s*;", re.DOTALL)),
                    validate.get("json"),
                    validate.parse_json(),
                    {
                        "type": str,
                        "id": int,
                    },
                    validate.union_get("type", "id"),
                ),
            ),
        )
        if not data:
            return

        vod_type, self.id = data
        if vod_type == "NEWS":
            self.id, self.title = self.session.http.get(
                self._URL_INFO_API_NEWS.format(id=self.id),
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "data": {
                            "video": {
                                "title": str,
                                "_id": int,
                            },
                        },
                    },
                    validate.get(("data", "video")),
                    validate.union_get("_id", "title"),
                ),
            )
        elif vod_type == "SG":
            self.id = self.session.http.get(
                self._URL_INFO_API_BLOCK_LIST.format(id=self.id),
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "data": {
                            "items": validate.all(
                                [
                                    {
                                        "_id": int,
                                        "type": str,
                                    },
                                ],
                                validate.filter(lambda obj: obj["type"] == "LIVE"),
                                validate.get(0),
                                validate.none_or_all(validate.get("_id")),
                            ),
                        },
                    },
                    validate.get(("data", "items")),
                ),
            )
            if not self.id:
                return
            self.id = self.session.http.get(
                self._URL_INFO_API_BLOCK.format(id=self.id),
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "data": {
                            "live_items": validate.all(
                                [
                                    {
                                        "_id": int,
                                        "type": str,
                                    },
                                ],
                                validate.filter(lambda obj: obj["type"] == "VIDEO"),
                                validate.get(0),
                                validate.none_or_all(validate.get("_id")),
                            ),
                        },
                    },
                    validate.get(("data", "live_items")),
                ),
            )
            if not self.id:
                return

        yield from self._get_formats_from_api(self.id)

    def _get_streams(self):
        if self.matches["live"]:
            return self._get_vod(self.match["channel_id"], "LIVE")
        if self.matches["vod"]:
            return self._get_vod(self.match["vod_id"], "MOVIE")
        if self.matches["tvp_info"]:
            return self._get_tvp_info()
        if self.matches["tvp_sport"]:
            return self._get_formats_from_api(self.match["stream_id"])

        return None


__plugin__ = TVP



class StreamlinkCLIError(Exception):
    def __init__(self, *args, code=1):
        super().__init__(*args)
        self.code = code

    def __str__(self):
        return super().__str__() or str(self.__context__ or "")



"""
$description Live TV channels and video on-demand service from ARD, a German public, independent broadcaster.
$url ardmediathek.de
$url mediathek.daserste.de
$type live, vod
$metadata title
$region Germany
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(\w+\.)?ardmediathek\.de/live/(?:[^/]+/)?(?P<id_live>\w+)(?:\?|$)"),
)
@pluginmatcher(
    name="video",
    pattern=re.compile(r"https?://(\w+\.)?ardmediathek\.de/video/(?:[^/]+/[^/]+/[^/]+/)?(?P<id_video>\w+)(?:\?|$)"),
)
class ARDMediathek(Plugin):
    _URL_API = "https://api.ardmediathek.de/page-gateway/pages/ard/item/{item}"
    _QUALITY_MAP = {
        4: "1080p",
        3: "720p",
        2: "540p",
        1: "360p",
        0: "270p",
    }

    def _get_streams(self):
        data_json = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[@type='application/json'][@id='fetchedContextValue2'][1]/text()"),
                validate.none_or_all(
                    validate.parse_json(),
                    [validate.list(str, {"data": dict})],
                    validate.filter(lambda item: item[0].startswith("https://api.ardmediathek.de/page-gateway/pages/")),
                    validate.any(
                        validate.get((0, 1, "data")),
                        [],
                    ),
                ),
            ),
        )
        if not data_json:
            data_json = self.session.http.get(
                self._URL_API.format(item=self.match["id_live"] if self.matches["live"] else self.match["id_video"]),
                params={
                    "devicetype": "pc",
                    "embedded": "false",
                },
                schema=validate.Schema(validate.parse_json()),
            )
        if not data_json:
            return

        schema_data = validate.Schema({
            "id": str,
            "widgets": validate.all(
                [dict],
                validate.filter(lambda item: item.get("mediaCollection")),
                validate.get(0),
                validate.any(
                    None,
                    validate.all(
                        {
                            "geoblocked": bool,
                            "publicationService": {
                                "name": str,
                            },
                            "show": validate.any(
                                None,
                                validate.all(
                                    {"title": str},
                                    validate.get("title"),
                                ),
                            ),
                            "title": str,
                            "mediaCollection": {
                                "embedded": {
                                    "_mediaArray": [
                                        validate.all(
                                            {
                                                "_mediaStreamArray": [
                                                    validate.all(
                                                        {
                                                            "_quality": validate.any(str, int),
                                                            "_stream": validate.url(),
                                                        },
                                                        validate.union_get("_quality", "_stream"),
                                                    ),
                                                ],
                                            },
                                            validate.get("_mediaStreamArray"),
                                            validate.transform(dict),
                                        ),
                                    ],
                                },
                            },
                        },
                        validate.union_get(
                            "geoblocked",
                            ("mediaCollection", "embedded", "_mediaArray", 0),
                            ("publicationService", "name"),
                            "title",
                            "show",
                        ),
                    ),
                ),
            ),
        })
        data = schema_data.validate(data_json)

        log.debug(f"Found media id: {data['id']}")
        if not data["widgets"]:
            log.info("The content is unavailable")
            return

        geoblocked, media, self.author, self.title, show = data["widgets"]
        if geoblocked:
            log.info("The content is not available in your region")
            return
        if show:
            self.title = f"{show}: {self.title}"

        if media.get("auto"):
            yield from HLSStream.parse_variant_playlist(self.session, media.get("auto")).items()
        else:
            for quality, stream in media.items():
                yield self._QUALITY_MAP.get(quality, quality), HTTPStream(self.session, stream)


__plugin__ = ARDMediathek



"""
$description Japanese live-streaming and video hosting social platform.
$url openrec.tv
$type live, vod
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?openrec\.tv/(?:live|movie)/(?P<id>[^/]+)"),
)
@pluginargument(
    "email",
    requires=["password"],
    metavar="EMAIL",
    help="The email associated with your openrectv account, required to access any openrectv stream.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="An openrectv account password to use with --openrectv-email.",
)
class OPENRECtv(Plugin):
    _stores_re = re.compile(r"window.stores\s*=\s*({.*?});", re.DOTALL | re.MULTILINE)
    _config_re = re.compile(r"window.sharedConfig\s*=\s*({.*?});", re.DOTALL | re.MULTILINE)

    movie_info_url = "https://public.openrec.tv/external/api/v5/movies/{id}"
    subscription_info_url = "https://apiv5.openrec.tv/api/v5/movies/{id}/detail"
    login_url = "https://www.openrec.tv/viewapp/v4/mobile/user/login"

    _info_schema = validate.Schema({
        validate.optional("id"): str,
        validate.optional("title"): str,
        validate.optional("movie_type"): str,
        validate.optional("onair_status"): validate.any(None, int),
        validate.optional("public_type"): str,
        validate.optional("media"): {
            "url": validate.any(None, validate.url()),
            "url_public": validate.any(None, validate.url()),
            "url_ull": validate.any(None, validate.url()),
        },
        validate.optional("subs_trial_media"): {
            "url": validate.any(None, validate.url()),
            "url_ull": validate.any(None, validate.url()),
        },
    })

    _subscription_schema = validate.Schema({
        validate.optional("status"): int,
        validate.optional("data"): {
            "items": [
                {
                    "media": {
                        "url": validate.any(None, validate.url()),
                    },
                },
            ],
        },
    })

    _login_schema = validate.Schema({
        validate.optional("error_message"): str,
        "status": int,
        validate.optional("data"): object,
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.video_id = None

    def login(self, email, password):
        data = self.session.http.post(
            self.login_url,
            data={"mail": email, "password": password},
            schema=validate.Schema(
                validate.parse_json(),
                self._login_schema,
            ),
        )
        if data["status"] == 0:
            log.debug("Logged in as %s", data["data"]["user_name"])
        else:
            log.error("Failed to login: %s", data["error_message"])
        return data["status"] == 0

    def _get_movie_data(self):
        url = self.movie_info_url.format(id=self.video_id)
        res = self.session.http.get(
            url,
            headers={
                "access-token": self.session.http.cookies.get("access_token"),
                "uuid": self.session.http.cookies.get("uuid"),
            },
        )
        data = self.session.http.json(res, schema=self._info_schema)

        if data["id"]:
            log.debug("Got valid detail response")
            return data
        else:
            log.error("Failed to get video stream: %s", data["message"])

    def _get_subscription_movie_data(self):
        url = self.subscription_info_url.format(id=self.video_id)
        res = self.session.http.get(
            url,
            headers={
                "access-token": self.session.http.cookies.get("access_token"),
                "uuid": self.session.http.cookies.get("uuid"),
            },
        )
        data = self.session.http.json(res, schema=self._subscription_schema)

        if data["status"] == 0:
            log.debug("Got valid subscription info")
            return data
        else:
            log.error("Failed to get video stream: %s", data["message"])

    def get_author(self):
        mdata = self._get_movie_data()
        if mdata:
            return mdata["channel"]["name"]

    def get_title(self):
        mdata = self._get_movie_data()
        if mdata:
            return mdata["title"]

    def _get_streams(self):
        self.video_id = self.url.rsplit("/", 1)[-1]
        if self.get_option("email") and self.get_option("password"):
            self.login(self.get_option("email"), self.get_option("password"))
        mdata = self._get_movie_data()

        if mdata:
            log.debug("Found video: %s (%s)", mdata["title"], mdata["id"])
            m3u8_file = None
            # subscription
            if mdata["public_type"] == "member":
                subs_data = self._get_subscription_movie_data()
                m3u8_file = subs_data["data"]["items"][0]["media"]["url"]
            # streaming
            elif mdata["onair_status"] == 1:
                m3u8_file = mdata["media"]["url_ull"] or mdata["subs_trial_media"]["url_ull"]
            # archive
            elif mdata["onair_status"] == 2 and mdata["media"]["url_public"] is not None:
                m3u8_file = mdata["media"]["url_public"].replace("public.m3u8", "playlist.m3u8")
            # uploaded video
            elif mdata["onair_status"] is None and mdata["movie_type"] == "2":
                m3u8_file = mdata["media"]["url"]
            else:
                log.error("There is no video file.")

            if m3u8_file is not None:
                yield from HLSStream.parse_variant_playlist(
                    self.session,
                    m3u8_file,
                    headers={"Referer": self.url},
                ).items()

        else:
            log.error("You don't have the authority or no video file.")


__plugin__ = OPENRECtv



class StreamlinkError(Exception):
    """
    Any error caused by Streamlink will be caught with this exception.
    """


# TODO: don't use PluginError for failed HTTP requests or validation schema failures
class PluginError(StreamlinkError):
    """
    Plugin related error.
    """


class FatalPluginError(PluginError):
    """
    Plugin related error that cannot be recovered from.

    Plugins should use this ``Exception`` when errors that can
    never be recovered from are encountered. For example, when
    a user's input is required and none can be given.
    """


class NoPluginError(StreamlinkError):
    """
    Error raised by :py:meth:`Streamlink.resolve_url() <streamlink.Streamlink.resolve_url()>`
    and :py:meth:`Streamlink.resolve_url_no_redirect() <streamlink.Streamlink.resolve_url_no_redirect()>`
    when no plugin could be found for the given input URL.
    """


class NoStreamsError(StreamlinkError):
    """
    Plugins should use this ``Exception`` in :py:meth:`Plugin._get_streams() <streamlink.plugin.Plugin._get_streams()>`
    when returning ``None`` or an empty ``dict`` is not possible, e.g. in nested function calls.
    """


class StreamError(StreamlinkError):
    """
    Stream related error.
    """


# https://stackoverflow.com/a/49797717
class _StreamlinkWarningMeta(type):
    def __new__(mcs, name, bases, namespace, **kw):
        name = namespace.get("__name__", name)
        return super().__new__(mcs, name, bases, namespace, **kw)


class StreamlinkWarning(UserWarning, metaclass=_StreamlinkWarningMeta):
    pass


class StreamlinkDeprecationWarning(StreamlinkWarning):
    __name__ = "StreamlinkDeprecation"


__all__ = [
    "StreamlinkError",
    "PluginError",
    "FatalPluginError",
    "NoPluginError",
    "NoStreamsError",
    "StreamError",
    "StreamlinkWarning",
    "StreamlinkDeprecationWarning",
]
